# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pydefect.analyzer.band_edge_states import PerfectBandEdgeState, EdgeInfo, \
    OrbitalInfo
from pydefect.cli.vasp.make_band_edge_orbital_infos import \
    calc_orbital_character
from pymatgen.electronic_structure.core import Spin
from vise.defaults import defaults as v_defaults
from pymatgen.io.vasp import Procar, Vasprun, Outcar
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties


def make_perfect_band_edge_state_from_vasp(
        procar: Procar, vasprun: Vasprun, outcar: Outcar
) -> PerfectBandEdgeState:

    band_edge_prop = VaspBandEdgeProperties(vasprun, outcar,
                                            v_defaults.integer_criterion)
    orbs, s = procar.data, vasprun.final_structure
    vbm_info = get_edge_info(band_edge_prop.vbm_info, orbs, s, vasprun)
    cbm_info = get_edge_info(band_edge_prop.cbm_info, orbs, s, vasprun)
    return PerfectBandEdgeState(vbm_info, cbm_info)


def get_edge_info(ei, orbs, s, vasprun):
    orbitals = calc_orbital_character(
        orbs, s, Spin.up, ei.kpoint_index, ei.band_index)
    e, occ = vasprun.eigenvalues[Spin.up][ei.kpoint_index, ei.band_index, :]
    orb_info = OrbitalInfo(energy=e, occupation=occ, orbitals=orbitals)
    return EdgeInfo(ei.band_index, tuple(ei.kpoint_coords), orb_info)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from itertools import product
from typing import List

import numpy as np
from pydefect.cli.vasp.make_efnv_correction import calc_max_sphere_radius
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Chgcar


class RadialDist:
    def __init__(self, parchg: Chgcar, center_coords: List[float]):
        self.data = parchg.spin_data
        self.dim = parchg.dim
        self.lattice = parchg.structure.lattice
        self.center = center_coords
        self.radius = calc_max_sphere_radius(self.lattice.matrix)
        self._distances_data = None  # lazy evaluation

    @property
    def distances_data(self):
        if self._distances_data is None:
            grid_points = [[x / self.dim[0], y / self.dim[1], z / self.dim[2]]
                           for (x, y, z) in
                           product(*[list(range(i)) for i in self.dim])]

            # Use boolean indexing to find charges within the desired distance.
            # data[:, 0]: shifted_coords
            # data[:, 1]: distances
            # data[:, 2]: sequential indices
            # data[:, 3]: images
            self._distances_data = np.array(self.lattice.get_points_in_sphere(
                grid_points, self.center, self.radius))

        return self._distances_data

    def histogram(self, spin: Spin, nbins: int = 15):
        inds = self.distances_data[:, 1] <= self.radius
        dists = self.distances_data[inds, 1]
        data_inds = np.rint(np.mod(list(self.distances_data[inds, 0]), 1) *
                            np.tile(self.dim, (len(dists), 1))).astype(int)

        vals = [self.data[spin][x, y, z] for x, y, z in data_inds]

        hist, edges = np.histogram(dists, nbins,
                                   range=[0, self.radius], weights=vals)
        hist_numbers, _ = np.histogram(dists, nbins, range=[0, self.radius])
        mesh_distance = edges[1] - edges[0]

        hist_data = np.zeros((nbins, 2))
        hist_data[:, 0] = [sum(edges[i:i + 2]) / 2 for i in range(nbins)]

        # 4pi * r^2 * rho
        integrated_volume = 4 * np.pi * hist_data[:, 0] ** 2
        density = hist / hist_numbers / self.lattice.volume
        hist_data[:, 1] = integrated_volume * density

        half_point = None
        for i in range(hist_data[:, 1].size):
            if sum(hist_data[:i + 1, 1]) > 0.5:
                # Obtain from the calculation of the area of a trapezoid
                x0 = hist_data[i - 1, 0]
                y0 = hist_data[i, 1]
                y1 = hist_data[i + 1, 1]
                half_point = (0.5 - sum(hist_data[:i, 1]) * mesh_distance) \
                    * 2 / (y0 + y1) + x0
                break

        summed = sum(hist_data[:, 1]) * mesh_distance
        return hist_data, half_point, summed





import joblib
import pytest
from pydefect.analyzer.grids import Grids
from pymatgen.core import Lattice, Structure
import numpy as np
from pymatgen.io.vasp import Chgcar
from vise.tests.helpers.assertion import assert_dataclass_almost_equal


@pytest.fixture
def grids():
    return Grids(lattice=Lattice.cubic(10),
                 dim=(1, 1, 5),
                 distance_data=np.array([[[0.0, 2.0, 4.0, 4.0, 2.0]]]))


@pytest.fixture
def chgcar():
    struc = Structure(lattice=Lattice.cubic(10), species=["H"], coords=[[0]*3])
    data = {"total": np.array([[[0.0, 2.0, 4.0, 6.0, 8.0]]]),
            "diff": np.array([[[0.0, 1.0, 2.0, 3.0, 4.0]]])}
    return Chgcar(struc, data=data)


def test_grids_joblib_roundtrip(tmpdir, grids):
    tmpdir.chdir()
    print(tmpdir)
    with open("tmp.joblib", mode="wb") as f:
        joblib.dump(grids, f, compress=3)
    with open("tmp.joblib", mode="rb") as f:
        actual = joblib.load(f)
    assert_dataclass_almost_equal(actual, grids)


def test_grids_np_save_load_roundtrip(tmpdir, grids):
    tmpdir.chdir()
    print(tmpdir)
    grids.dump()
    actual = grids.from_file()
    assert_dataclass_almost_equal(actual, grids)


def test_grids_from_chgcar(grids, chgcar):
    actual = Grids.from_chgcar(chgcar)
    assert_dataclass_almost_equal(actual, grids)


def test_shift_distance_data(grids):
    actual = grids.shifted_distance_data(center=[0, 0, 1])
    expected = np.array([[[2.0, 0.0, 2.0, 4.0, 4.0]]])
    np.testing.assert_array_almost_equal(actual, expected)


def test_shift_distance_data2():
    grids = Grids(lattice=Lattice.cubic(10),
                  dim=(2, 2, 2),
                  distance_data=np.array([[[0.0, 5.0], [5.0, 7.07]],
                                          [[5.0, 7.07], [7.07, 8.66]]]))
    actual = grids.shifted_distance_data(center=[1, 1, 1])
    expected = np.array([[[8.66, 7.07], [7.07, 5.0]],
                         [[7.07, 5.0], [5.0, 0.0]]])
    np.testing.assert_array_almost_equal(actual, expected)


def test_spherical_dist(grids):
    # distance_data=np.array([[[0.0, 2.0, 4.0, 4.0, 2.0]]]))
    actual = grids.spherical_dist(data=np.array([[[0.0, 0.0, 0.0, 0.0, 1.0]]]),
                                  center=[0, 0, 1],
                                  distance_bins=np.array([0.0, 2.5, 5.0]))
    #  Divide by 2 since there are 2 points at 4.0 distance. volume=1000.
    expected = [0.0, 1.0 / 2 / 1000]
    assert actual == expected


"""
TODO
- Check how to revolve numpy array.
- Add distances_data
- Add _calc_histogram(chgcar, distances_data, center) method

DONE
- Add defect_center_idxs
- Add defect_center_coords property
"""


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import pytest
from pydefect.input_maker.local_extrema import VolumetricDataLocalExtrema, \
    CoordInfo
from pydefect.input_maker.supercell_info import Interstitial, SupercellInfo
from pydefect.util.structure_tools import Coordination
from pymatgen.core import Structure
from vise.tests.helpers.assertion import assert_json_roundtrip
from vise.util.structure_symmetrizer import Site


@pytest.fixture
def local_extrema(simple_cubic, vol_params):
    coordination = Coordination({"Mn": [1.0, 2.0]}, cutoff=4.0,
                                neighboring_atom_indices=[1, 2])
    local_extremum = CoordInfo(site_symmetry="3m",
                               coordination=coordination,
                               frac_coords=[(0.1, 0.1, 0.1)],
                               quantities=[1.234])

    coordination = Coordination({"Mn": [1.0, 2.0]}, cutoff=4.0,
                                neighboring_atom_indices=[1, 2])
    local_extremum2 = CoordInfo(site_symmetry="4mm",
                                coordination=coordination,
                                frac_coords=[(0.1, 0.0, 0.0)],
                                quantities=[2.101])

    return VolumetricDataLocalExtrema(unit_cell=simple_cubic, is_min=True,
                                      extrema_points=[local_extremum,
                                                      local_extremum2],
                                      info="test",
                                      params=vol_params)


def test_local_extrema_json_roundtrip(local_extrema, tmpdir):
    tmpdir.chdir()
    assert_json_roundtrip(local_extrema, tmpdir)


def test_local_extrema_str(local_extrema, tmpdir):
    actual = local_extrema.__str__()
    expected = """info: test
min_or_max: min
extrema_points:
#  site_sym  coordination        frac_coords               quantity
1  3m        {'Mn': [1.0, 2.0]}  ( 0.100,  0.100,  0.100)  1.2
2  4mm       {'Mn': [1.0, 2.0]}  ( 0.100,  0.000,  0.000)  2.1"""
    print(actual)
    assert actual == expected


def test_local_extrema_append_sites_to_supercell_info(local_extrema,
                                                      simple_cubic):
    sites = {"H1": Site(element="H", wyckoff_letter="a", site_symmetry="mmm",
                        equivalent_atoms=[0, 1, 2, 3, 4, 5, 6, 7])}
    actual_cubic = Structure.from_dict(simple_cubic.as_dict())
    supercell_info = SupercellInfo(actual_cubic * [2, 2, 2],
                                   "Pm-3m",
                                   [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
                                   sites, [])
    actual = local_extrema.append_sites_to_supercell_info(supercell_info, indices=[1, 2])

    interstitials = [Interstitial(frac_coords=[0.05]*3, site_symmetry="3m", info="test #1"),
                     Interstitial(frac_coords=[0.05, 0.0, 0.0], site_symmetry="4mm", info="test #2")]
    expected = SupercellInfo(actual_cubic * [2, 2, 2],
                             "Pm-3m",
                             [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
                             sites, interstitials)
    assert actual == expected


def test_local_extrema_append_sites_to_supercell_info2(local_extrema,
                                                       simple_cubic):
    sites = {"H1": Site(element="H", wyckoff_letter="a", site_symmetry="mmm",
                        equivalent_atoms=[0, 1, 2, 3, 4, 5, 6, 7])}
    actual_cubic = Structure.from_dict(simple_cubic.as_dict())
    supercell_info = SupercellInfo(actual_cubic * [2, 2, 2],
                                   "Pm-3m",
                                   [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
                                   sites, [])

    actual_2 = local_extrema.append_sites_to_supercell_info(supercell_info, indices=[2])

    interstitials = [Interstitial(frac_coords=[0.05, 0.0, 0.0], site_symmetry="4mm", info="test #2")]
    expected_2 = SupercellInfo(actual_cubic * [2, 2, 2],
                             "Pm-3m",
                             [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
                             sites, interstitials)
    assert actual_2 == expected_2



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import numpy as np
import pytest
from pydefect.analyzer.defect_charge_info import AveChargeDensityDist, \
    DefectChargeInfo
from pydefect.analyzer.grids import Grids
from pydefect.cli.vasp.make_defect_charge_info import \
    center_1d_periodic_quantity, make_charge_dist, make_defect_charge_info
from pymatgen.core import Structure, Lattice
from pymatgen.io.vasp import Chgcar


def test_center_1d_periodic_quantity():
    grid_points = [0.3, 0.2, 0.1, 0.0, 0.1, 0.2, 0.3, 0.4]
    assert center_1d_periodic_quantity(grid_points) == 7
    grid_points = [0.3, 0.2, 0.1, 0.1, 0.2, 0.3, 0.4]
    assert center_1d_periodic_quantity(grid_points) == 6


@pytest.fixture
def parchg():
    struc = Structure(lattice=Lattice.cubic(10), species=["H"], coords=[[0]*3])
    data = {"total": np.array([[[0, 0], [0, 2000.0]], [[0, 0], [0, 0]]]),
            "diff": np.array([[[0, 0], [0, 1000.0]], [[0, 0], [0, 0]]])}
    # spin-up: 1.5,  spin-down: 0.5
    return Chgcar(struc, data=data)


@pytest.fixture
def charge_dist_list():
    return [AveChargeDensityDist((0.0, 0.5, 0.5), [1.5, 0.0]),
            AveChargeDensityDist((0.0, 0.5, 0.5), [0.5, 0.0])]


def test_make_charge_dist(parchg, charge_dist_list):
    grids = Grids(lattice=Lattice.cubic(10),
                  dim=(2, 2, 2),
                  distance_data=np.array(
                      [[[0., 5.], [5., 7.07106781]],
                       [[5., 7.07106781], [7.07106781, 8.66025404]]]))
    bins = np.array([0.0, 2.5, 5.0])
    actual = make_charge_dist(parchg=parchg, grids=grids, distance_bins=bins)
    # divide by 3 as there are three points in the second bin.
    assert actual == charge_dist_list


def test_make_defect_charge_info(parchg, charge_dist_list):
    grids = Grids.from_chgcar(parchg)
    actual = make_defect_charge_info([parchg], band_idxs=[1], bin_interval=2.5,
                                     grids=grids)
    expected = DefectChargeInfo([0.0, 2.5, 5.0],
                                [1],
                                [charge_dist_list],
                                0.001)
    assert actual == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy

import pytest

from pydefect.analyzer.defect_energy import (
    DefectEnergy, CrossPoints, DefectEnergyInfo, DefectEnergies,
    DefectEnergySummary, SingleChargeEnergies, ChargeEnergies)
from vise.tests.helpers.assertion import assert_yaml_roundtrip, \
    assert_json_roundtrip


@pytest.fixture
def defect_energy():
    return DefectEnergy(formation_energy=1.0,
                        energy_corrections={"1st order": 1.0, "alignment": 2.0},
                        is_shallow=False)


def test_energy_total_correction(defect_energy):
    assert defect_energy.total_correction == 3.0
    assert defect_energy.energy(with_correction=True) == 4.0


@pytest.fixture
def defect_energy_info(defect_energy):
    return DefectEnergyInfo(name="Va_O1", charge=1, atom_io={"O": -1},
                            defect_energy=defect_energy)


@pytest.fixture
def defect_energy_info2():
    energy = DefectEnergy(formation_energy=0.0,
                          energy_corrections={"no correction": 0.0})
    return DefectEnergyInfo(name="hole polaron", charge=1, atom_io={},
                            defect_energy=energy)


def test_energy_yaml(defect_energy_info, tmpdir):
    expected_text = """name: Va_O1
charge: 1
formation_energy: 1.0
atom_io:
  O: -1
energy_corrections:
  1st order: 1.0
  alignment: 2.0
is_shallow: False"""
    assert_yaml_roundtrip(defect_energy_info, tmpdir, expected_text)


def test_energy_yaml2(defect_energy_info2, tmpdir):
    expected_text = """name: hole polaron
charge: 1
formation_energy: 0.0
atom_io:
energy_corrections:
  no correction: 0.0
is_shallow: """
    assert_yaml_roundtrip(defect_energy_info2, tmpdir, expected_text)


@pytest.fixture
def defect_energies():
    return DefectEnergies(
        atom_io={"O": -1},
        charges=[0, 1, 2],
        defect_energies=[DefectEnergy(1.0, {"corr": 2.0}, is_shallow=False),
                         DefectEnergy(2.0, {"corr": 2.0}, is_shallow=False),
                         DefectEnergy(3.0, {"corr": 2.0}, is_shallow=True)])


@pytest.fixture
def defect_energy_summary(defect_energies):
    return DefectEnergySummary(
        title="MgAl2O4",
        defect_energies={"Va_O1": defect_energies},
        rel_chem_pots={"A": {"O": -1.0}, "B": {"O": -2.0}},
        cbm=2.0, supercell_vbm=-1.0, supercell_cbm=3.0)


def test_defect_energy_summary_json_roundtrip(defect_energy_summary, tmpdir):
    assert_json_roundtrip(defect_energy_summary, tmpdir)


def test_defect_energy_summary_str():
    defect_energies = DefectEnergies(
        atom_io={"O": -1},
        charges=[1, 2, 0],
        defect_energies=[
            DefectEnergy(2.0001, {"corr": 2.0001}, is_shallow=False),
            DefectEnergy(3.0001, {"corr": 2.0001}, is_shallow=True),
            DefectEnergy(1.0001, {"corr": 2.0001}, is_shallow=False),
        ])

    defect_energy_summary = DefectEnergySummary(
        title="MgAl2O4",
        defect_energies={"Va_O1": defect_energies},
        rel_chem_pots={"A": {"O": -1.000000001}, "B": {"O": -2.000000001}},
        cbm=2.000000001, supercell_vbm=-1.000000001, supercell_cbm=3.000000001)

    actual = defect_energy_summary.__str__()
    expected = """title: MgAl₂O₄
rel_chem_pots:
 -A O: -1.00
 -B O: -2.00
vbm: 0.00, cbm: 2.00, supercell vbm: -1.00, supercell_cbm: 3.00

name    atom_io      charge    energy    correction  is_shallow
------  ---------  --------  --------  ------------  ------------
Va_O1   O: -1             0     1.000         2.000  False
                          1     2.000         2.000  False
                          2     3.000         2.000  True"""
    assert actual == expected


def test_defect_energy_summary_charge_energies(
        defect_energy_summary, defect_energies):
    actual = defect_energy_summary.charge_energies(
        chem_pot_label="A", allow_shallow=True, with_corrections=False, e_range=(0.0, 2.0))
    expected = ChargeEnergies(
        charge_energies_dict={"Va_O": SingleChargeEnergies([(0, 0.0), (1, 1.0), (2, 2.0)])},
        e_min=0.0, e_max=2.0)
    assert actual == expected

    # change chem_pot_label
    actual = defect_energy_summary.charge_energies(
        chem_pot_label="B", allow_shallow=True, with_corrections=False, e_range=(0.0, 2.0))
    expected = ChargeEnergies(
        charge_energies_dict={"Va_O": SingleChargeEnergies([(0, -1.0), (1, 0.0), (2, 1.0)])},
        e_min=0.0, e_max=2.0)
    assert actual == expected

    # change allow_shallow
    actual = defect_energy_summary.charge_energies(
        chem_pot_label="A", allow_shallow=False, with_corrections=False, e_range=(0.0, 2.0))
    expected = ChargeEnergies(
        charge_energies_dict={"Va_O": SingleChargeEnergies([(0, 0.0), (1, 1.0)])},
        e_min=0.0, e_max=2.0)
    assert actual == expected

    # change with_correction
    actual = defect_energy_summary.charge_energies(
        chem_pot_label="A", allow_shallow=True, with_corrections=True, e_range=(0.0, 2.0))
    expected = ChargeEnergies(
        charge_energies_dict={"Va_O": SingleChargeEnergies([(0, 2.0), (1, 3.0), (2, 4.0)])},
        e_min=0.0, e_max=2.0)
    assert actual == expected


def test_cross_points_dicts():
    charge_energies = ChargeEnergies(
        charge_energies_dict={"Va_Mg": SingleChargeEnergies([(0, 2.0)]),
                              "Va_O": SingleChargeEnergies([(1, 1.0)])},
        e_min=0.0, e_max=2.0)
    actual = charge_energies.cross_point_dicts
    expected = {"Va_Mg": CrossPoints(inner_cross_points=[],
                                     boundary_points=[[0.0, 2.0], [2.0, 2.0]]),
                "Va_O": CrossPoints(inner_cross_points=[],
                                    boundary_points=[[0.0, 1.0], [2.0, 3.0]])}
    assert actual == expected


def test_defect_energy_summary_latexified_title(defect_energy_summary):
    actual = defect_energy_summary.latexified_title
    expected = "MgAl$_{2}$O$_{4}$"
    assert actual == expected


@pytest.fixture
def single_charge_energies():
    return SingleChargeEnergies([(0, 6.0), (1, 3.0)])


@pytest.fixture
def charge_energies(single_charge_energies):
    return ChargeEnergies({"Va_O": single_charge_energies},
                          e_min=1.0, e_max=6.0)


def test_charge_energies_cross_points(charge_energies):
    actual = charge_energies.cross_point_dicts
    expected = {"Va_O": CrossPoints([[3.0, 6.0]], [[6.0, 6.0], [1.0, 4.0]])}
    assert actual == expected


def test_charge_energies_energy_range(charge_energies):
    actual = charge_energies.energy_range(0.2)
    expected = [3.8, 6.2]
    assert actual == expected


def test_charge_energies_transition_levels(single_charge_energies):
    actual = single_charge_energies.transition_levels
    expected = {(0, 1): 3.0}
    assert actual == expected


def test_charge_energies_pinning_level(single_charge_energies):
    assert single_charge_energies.pinning_level(1.0, 6.0) == (None, None)
    ce = SingleChargeEnergies(charge_energies=[(0, 6.0), (1, 3.0)])
    assert ce.pinning_level(-4.0, 6.0) == ((-3.0, 1), None)


def test_charge_energies_energy_at_ef(single_charge_energies):
    assert single_charge_energies.energy_at_ef(ef=0.0) == (3.0, 1)
    assert single_charge_energies.energy_at_ef(ef=10.0) == (6.0, 0)


@pytest.fixture
def cross_points():
    inner_cross_points = [[2, 30], [3, 40]]
    boundary_points = [[1, 10], [4, 40]]
    return CrossPoints(inner_cross_points, boundary_points)


def test_cross_points(cross_points):
    assert cross_points.all_sorted_points == [[1, 10], [2, 30], [3, 40], [4, 40]]
    assert cross_points.t_all_sorted_points == [[1, 2, 3, 4], [10, 30, 40, 40]]
    assert cross_points.t_inner_cross_points == [[2, 3], [30, 40]]
    assert cross_points.t_boundary_points == [[1, 4], [10, 40]]
    assert cross_points.charges == [20, 10, 0]
    assert cross_points.charge_list == [(None, 20), (20, 10), (10, 0), (0, None)]
    assert cross_points.annotated_charge_positions == \
           {20: [1.5, 20], 10: [2.5, 35], 0: [3.5, 40]}


def test_cross_points_str(cross_points):
    expected = """      1.0000      10.0000
      2.0000      30.0000
      3.0000      40.0000
      4.0000      40.0000"""
    assert str(cross_points) == expected




from typing import List, Tuple

from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.calc_summary import CalcSummary, SingleCalcSummary
from pydefect.analyzer.defect_structure_info import DefectStructureInfo
from pydefect.analyzer.make_defect_energy_info import num_atom_differences
from pydefect.defaults import defaults
from pydefect.input_maker.defect_entry import DefectEntry


def make_calc_summary(
        calc_set: List[Tuple[CalcResults, DefectEntry, DefectStructureInfo]],
        p_calc_results: CalcResults) -> CalcSummary:

    summaries = {}
    for calc_results, entry, str_info in calc_set:
        summaries[entry.full_name] = \
            make_single_calc_summary(calc_results, entry, p_calc_results,
                                     str_info)
    return CalcSummary(single_summaries=summaries)


def make_single_calc_summary(calc_results, entry, p_calc_results, str_info):
    atom_io = num_atom_differences(calc_results.structure,
                                   p_calc_results.structure)
    relative_energy = calc_results.energy - p_calc_results.energy
    is_energy_strange = abs(relative_energy) > defaults.abs_strange_energy
    return SingleCalcSummary(
        charge=entry.charge,
        atom_io=atom_io,
        electronic_conv=calc_results.electronic_conv,
        ionic_conv=calc_results.ionic_conv,
        is_energy_strange=is_energy_strange,
        same_config_from_init=str_info.same_config_from_init,
        defect_type=str(str_info.defect_type),
        symm_relation=str(str_info.symm_relation))




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

import pytest
from monty.serialization import loadfn
from pydefect.analyzer.band_edge_states import PerfectBandEdgeState, \
    BandEdgeOrbitalInfos
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.unitcell import Unitcell
from pydefect.chem_pot_diag.chem_pot_diag import CompositionEnergies, \
    CompositionEnergy, RelativeEnergies, ChemPotDiag
from pydefect.cli.main_functions import make_standard_and_relative_energies, \
    make_cpd_and_vertices, plot_chem_pot_diag, make_supercell, \
    append_interstitial_to_supercell_info, \
    pop_interstitial_from_supercell_info, make_defect_set, \
    make_band_edge_states_main_func, make_efnv_correction_main_func, \
    calc_defect_structure_info
from pydefect.corrections.efnv_correction import ExtendedFnvCorrection
from pydefect.input_maker.defect import SimpleDefect
from pydefect.input_maker.defect_entry import DefectEntry
from pydefect.input_maker.defect_set import DefectSet
from pymatgen.core import Composition, IStructure


def test_make_standard_and_relative_energies(mocker, tmpdir):
    tmpdir.chdir()
    args = Namespace(composition_energies_yaml="a.yaml")
    mock = mocker.patch("pydefect.cli.main_functions.CompositionEnergies")
    mock.from_yaml.return_value = CompositionEnergies(
        {Composition("H"): CompositionEnergy(-1.0),
         Composition("He"): CompositionEnergy(-2.0),
         Composition("HHe"): CompositionEnergy(-5.0)})
    make_standard_and_relative_energies(args)
    mock.from_yaml.assert_called_once_with("a.yaml")
    actual = Path("standard_energies.yaml").read_text()
    expected = """H: -1.0
He: -2.0
"""
    assert actual == expected

    actual = Path("relative_energies.yaml").read_text()
    expected = """HHe: -1.0
"""
    assert actual == expected


def test_make_cpd_and_vertices(mocker, tmpdir):
    tmpdir.chdir()
    mock = mocker.patch("pydefect.cli.main_functions.RelativeEnergies")
    mock.from_yaml.return_value = RelativeEnergies({"HHe": -1.0})

    args = Namespace(rel_energy_yaml="a.yaml", target="HHe", elements=None)
    make_cpd_and_vertices(args)

    actual = Path("target_vertices.yaml").read_text()
    expected = """target: HHe
A:
  chem_pot:
    H: 0.0
    He: -2.0
  competing_phases:
  - H
  impurity_phases: []
B:
  chem_pot:
    H: -2.0
    He: 0.0
  competing_phases:
  - He
  impurity_phases: []
"""
    assert actual == expected


def test_plot_chem_pot_diag(tmpdir):
    tmpdir.chdir()
    chem_pot_diag = ChemPotDiag(vertex_elements=["Mg", "O"],
                                polygons={"Mg": [[0.0, -3.0], [0.0, -1.0]],
                                          "O": [[-3.0, 0.0], [-1.0, 0.0]],
                                          "MgO": [[-1.0, 0.0], [0.0, -1.0]]})
    args = Namespace(chem_pot_diag=chem_pot_diag)
    plot_chem_pot_diag(args)


def test_make_supercell_from_matrix(simple_cubic, simple_cubic_2x1x1, tmpdir):
    tmpdir.chdir()
    matrix = [2, 1, 1]
    args = Namespace(unitcell=simple_cubic, matrix=matrix,
                     min_num_atoms=None, max_num_atoms=None)
    make_supercell(args)
    info = loadfn("supercell_info.json")
    assert IStructure.from_file("SPOSCAR") == simple_cubic_2x1x1
    assert info.structure == simple_cubic_2x1x1
    assert info.transformation_matrix == [[2, 0, 0], [0, 1, 0], [0, 0, 1]]


def test_make_recommended_supercell(simple_cubic, simple_cubic_2x2x2, tmpdir):
    tmpdir.chdir()
    args = Namespace(unitcell=simple_cubic, matrix=None,
                     min_num_atoms=8, max_num_atoms=8)
    make_supercell(args)
    info = loadfn("supercell_info.json")
    assert IStructure.from_file("SPOSCAR") == simple_cubic_2x2x2
    assert info.structure == simple_cubic_2x2x2
    assert info.transformation_matrix == [[2, 0, 0], [0, 2, 0], [0, 0, 2]]


def test_add_interstitials(mocker):
    mock_1 = mocker.Mock()
    mock_2 = mocker.Mock()
    mock_3 = mocker.Mock()
    mock_4 = mocker.Mock()
    args = Namespace(supercell_info=mock_1, base_structure=mock_2,
                     frac_coords=mock_3, info=mock_4)
    mock = mocker.patch("pydefect.cli.main_functions.append_interstitial")

    append_interstitial_to_supercell_info(args)
    mock.assert_called_once_with(mock_1, mock_2, [mock_3], mock_4)
    mock.return_value.to_json_file.assert_called_once_with()


def test_pop_interstitials(mocker):
    mock_si = mocker.MagicMock()
    args = Namespace(supercell_info=mock_si, index=1000, pop_all=False)

    pop_interstitial_from_supercell_info(args)
    assert isinstance(mock_si.interstitials, object)
    mock_si.interstitials.pop.assert_called_once_with(999)
    mock_si.to_json_file.assert_called_once_with()


def test_pop_interstitials_pop_all(mocker):
    mock_si = mocker.MagicMock()
    mock_si.interstitials = ["a", "b"]
    args = Namespace(supercell_info=mock_si, index=None, pop_all=True)

    pop_interstitial_from_supercell_info(args)
    assert isinstance(mock_si.interstitials, object)
    mock_si.interstitials = []


@pytest.mark.parametrize("oxi_states,he_vacancy_charge",
                         ([None, [0]], [["He", 1, "Li", 1], [-1, 0, 1]]))
def test_make_defect_set(oxi_states, he_vacancy_charge, tmpdir, supercell_info):
    tmpdir.chdir()
    supercell_info.to_json_file()
    args = Namespace(oxi_states=oxi_states, dopants=["Li"],
                     keywords=["Li_H1", "Va_He1", "Va_H1_-1"])
    make_defect_set(args)

    simple_defects = {SimpleDefect(None, "He1", he_vacancy_charge),
                      SimpleDefect(None, "H1", [-1]),
                      SimpleDefect("Li", "H1", [0])}

    DefectSet(defects=simple_defects).to_yaml("expected.yaml")
    actual = Path("defect_in.yaml").read_text()
    expected = Path("expected.yaml").read_text()
    assert actual == expected


def test_calc_defect_structure_info(mocker, tmpdir, supercell_info):
    tmpdir.chdir()
    supercell_info.to_json_file()
    mock_calc_results = mocker.Mock()
    mock_defect_entry = mocker.Mock()

    def side_effect(key):
        if str(key) == "Va_O1_2/calc_results.json":
            return mock_calc_results
        elif str(key) == "Va_O1_2/defect_entry.json":
            return mock_defect_entry
        else:
            print(key)
            raise ValueError

    mock_loadfn = mocker.patch(
        "pydefect.cli.main_functions.loadfn", side_effect=side_effect)
    mock_structure_info = mocker.patch(
        "pydefect.cli.main_functions.MakeDefectStructureInfo")
    info = loadfn("supercell_info.json")

    args = Namespace(supercell_info=info, check_calc_results=True,
                     dirs=[Path("Va_O1_2")], dist_tolerance=0.1, symprec=0.2,
                     verbose=False)
    calc_defect_structure_info(args)
    mock_structure_info.assert_called_with(
        supercell_info.structure,
        mock_defect_entry.structure,
        mock_calc_results.structure,
        dist_tol=0.1,
        symprec=0.2)
    mock_structure_info.return_value.defect_structure_info.to_json_file.assert_called_once_with(
        "Va_O1_2/defect_structure_info.json")


def test_make_efnv_correction_from_vasp(tmpdir, mocker):
    mock_defect_entry = mocker.Mock(spec=DefectEntry, autospec=True)
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)

    def side_effect(key):
        if str(key) == "Va_O1_2/defect_entry.json":
            mock_defect_entry.full_name = "Va_O1_2"
            mock_defect_entry.charge = 2
            return mock_defect_entry
        elif str(key) == "Va_O1_2/calc_results.json":
            return mock_calc_results
        else:
            raise ValueError

    mock_perfect_calc_results = mocker.Mock(spec=CalcResults)
    mock_loadfn = mocker.patch(
        "pydefect.cli.main_functions.loadfn", side_effect=side_effect)
    mock_unitcell = mocker.Mock(spec=Unitcell)
    mock_make_efnv = mocker.patch(
        "pydefect.cli.main_functions.make_efnv_correction")
    mock_efnv = mocker.Mock(spec=ExtendedFnvCorrection, autospec=True)
    mock_make_efnv.return_value = mock_efnv

    mock_site_pot_plotter = mocker.patch(
        "pydefect.cli.main_functions.SitePotentialMplPlotter")
    plotter = mock_site_pot_plotter.from_efnv_corr.return_value
    args = Namespace(dirs=[Path("Va_O1_2")],
                     check_calc_results=True,
                     perfect_calc_results=mock_perfect_calc_results,
                     unitcell=mock_unitcell,
                     verbose=False)

    make_efnv_correction_main_func(args)
    mock_loadfn.assert_any_call(Path("Va_O1_2") / "defect_entry.json")
    mock_loadfn.assert_any_call(Path("Va_O1_2") / "calc_results.json")

    mock_make_efnv.assert_called_with(
        mock_defect_entry.charge, mock_calc_results, mock_perfect_calc_results,
        mock_unitcell.dielectric_constant)
    mock_efnv.to_json_file.assert_called_with(
        Path("Va_O1_2") / "correction.json")

    mock_site_pot_plotter.from_efnv_corr.assert_called_with(
        title="Va_O1_2", efnv_correction=mock_efnv)
    plotter.construct_plot.assert_called_once_with()


def test_make_band_edge_states(mocker):
    mock_perfect_edge_states = mocker.Mock(
        spec=PerfectBandEdgeState, autospec=True)
    mocker_band_edge_orbital_infos = mocker.Mock(spec=BandEdgeOrbitalInfos,
                                                 autospec=True)

    def side_effect(key):
        if str(key) == "Va_O1_2/band_edge_orbital_infos.json":
            return mocker_band_edge_orbital_infos
        else:
            raise FileNotFoundError

    mocker_loadfn = mocker.patch(
        "pydefect.cli.main_functions.loadfn", side_effect=side_effect)
    mocker_make_edge_states = mocker.patch(
        "pydefect.cli.main_functions.make_band_edge_states")

    args = Namespace(dirs=[Path("Va_O1_2")],
                     check_calc_results=True,
                     p_state=mock_perfect_edge_states,
                     verbose=False)
    make_band_edge_states_main_func(args)
    mocker_make_edge_states.assert_called_with(mocker_band_edge_orbital_infos,
                                               mock_perfect_edge_states,
                                               None)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

from pydefect.cli.vasp.main_vasp_util import parse_args_main_vasp_util


def test_calc_charge_state():
    parsed_args = parse_args_main_vasp_util(
        ["ccs", "-d", "Va_O1_0"])
    expected = Namespace(
        dir=Path("Va_O1_0"),
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_defect_entry(mocker):
    mock_structure = mocker.patch("pydefect.cli.vasp.main_vasp_util.Structure")
    parsed_args = parse_args_main_vasp_util(
        ["de", "-d", "Va_O1_0", "-n", "Va_O1", "-p", "POSCAR"])
    expected = Namespace(
        dir=Path("Va_O1_0"),
        name="Va_O1",
        perfect=mock_structure.from_file.return_value,
        func=parsed_args.func)
    assert parsed_args == expected
    mock_structure.from_file.assert_called_once_with("POSCAR")


def test_make_parchg_dir():
    parsed_args = parse_args_main_vasp_util(
        ["pd", "-d", "Va_O1_0"])
    expected = Namespace(
        dir=Path("Va_O1_0"),
        func=parsed_args.func)
    assert parsed_args == expected


def test_refine_defect_poscar(mocker, simple_cubic):
    mock_structure = mocker.patch(
        "pydefect.cli.vasp.main_vasp_util.Structure")
    mock_loadfn = mocker.patch("pydefect.cli.vasp.main_vasp_util.loadfn")

    parsed_args = parse_args_main_vasp_util(
        ["rdp", "-p", "POSCAR", "-d", "defect_entry.json", "-n", "refined"])
    expected = Namespace(
        structure=mock_structure.from_file.return_value,
        defect_entry=mock_loadfn.return_value,
        poscar_name="refined",
        func=parsed_args.func)
    assert parsed_args == expected
    mock_structure.from_file.assert_called_once_with("POSCAR")
    mock_loadfn.assert_called_once_with("defect_entry.json")


def test_calc_grids(mocker):
    mock_chgcar = mocker.patch("pydefect.cli.vasp.main_vasp_util.Chgcar")
    parsed_args = parse_args_main_vasp_util(
        ["cg", "-c", "CHG"])
    expected = Namespace(
        chgcar=mock_chgcar.from_file.return_value,
        func=parsed_args.func)
    assert parsed_args == expected
    mock_chgcar.from_file.assert_called_once_with("CHG")


def test_calc_defect_charge_info(mocker):
    mock_grids = mocker.patch("pydefect.cli.vasp.main_vasp_util.Grids")
    parsed_args = parse_args_main_vasp_util(
        ["cdc", "-p", "PARCHG.0001.ALLK", "-b", "0.3", "-g", "Grids.npz"])
    expected = Namespace(
        parchgs=["PARCHG.0001.ALLK"],
        bin_interval=0.3,
        grids=mock_grids.from_file.return_value,
        func=parsed_args.func)
    assert parsed_args == expected
    mock_grids.from_file.assert_called_once_with("Grids.npz")





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import Optional, List, Tuple

from monty.json import MSONable
from pydefect.analyzer.defect_structure_comparator import \
    SiteDiff, SiteInfo
from pydefect.defaults import defaults
from pydefect.util.coords import pretty_coords
from pymatgen.core import Structure
from pymatgen.symmetry.groups import SpaceGroup
from tabulate import tabulate
from vise.util.enum import ExtendedEnum
from vise.util.logger import get_logger
from vise.util.mix_in import ToJsonFileMixIn

logger = get_logger(__name__)


class DefectType(MSONable, ExtendedEnum):
    vacancy = "vacancy"
    interstitial = "interstitial"
    substituted = "substituted"
    vacancy_split = "vacancy_split"
    interstitial_split = "interstitial_split"
    unknown = "unknown"


def judge_defect_type(site_diff: SiteDiff):
    if site_diff.is_vacancy:
        return DefectType.vacancy
    elif site_diff.is_interstitial:
        return DefectType.interstitial
    elif site_diff.is_substituted:
        return DefectType.substituted

    elements_involved = set()
    for _, elem, _ in site_diff.removed + site_diff.inserted:
        elements_involved.add(elem)

    if len(elements_involved) == 1 and not site_diff.removed_by_sub:
        if len(site_diff.removed) - len(site_diff.inserted) == 1:
            return DefectType.vacancy_split
        elif len(site_diff.removed) - len(site_diff.inserted) == -1:
            return DefectType.interstitial_split

    return DefectType.unknown


def remove_dot(x):
    return "".join([s for s in x if s != "."])


def unique_point_group(pg):
    result = remove_dot(pg)
    if result == "2mm" or result == "m2m":
        return "mm2"
    if result == "-4m2":
        return "-42m"
    if result == "m3":
        return "m-3"
    return result


class SymmRelation(MSONable, ExtendedEnum):
    same = "same"
    subgroup = "subgroup"
    supergroup = "supergroup"
    another = "another"


@dataclass
class Displacement(MSONable):
    specie: str
    original_pos: Tuple[float, float, float]
    final_pos: Tuple[float, float, float]
    distance_from_defect: float
    disp_vector: Tuple[float, float, float]
    displace_distance: float
    angle: Optional[float]


def symmetry_relation(initial_point_group, final_point_group):
    """ Check the point group symmetry relation using the space group relation
    implemented in pymatgen.
    """
    if initial_point_group in ["3m", "-3m"]:
        initial_point_group += "1"
    if final_point_group in ["3m", "-3m"]:
        final_point_group += "1"

    initial = SpaceGroup(f"P{initial_point_group}")
    final = SpaceGroup(f"P{final_point_group}")
    if initial == final:
        return SymmRelation.same
    elif final.is_subgroup(initial):
        return SymmRelation.subgroup
    elif final.is_supergroup(initial):
        return SymmRelation.supergroup
    else:
        return SymmRelation.another


@dataclass
class DefectStructureInfo(MSONable, ToJsonFileMixIn):
    shifted_final_structure: Structure
    initial_site_sym: str
    final_site_sym: str
    site_diff: SiteDiff
    site_diff_from_initial: SiteDiff
    symprec: float
    dist_tol: float
    anchor_atom_idx: int
    neighbor_atom_indices: List[int]
    neighbor_cutoff_factor: float
    drift_vector: Tuple[float, float, float]
    drift_dist: float
    center: Tuple[float, float, float]
    displacements: List[Displacement]

    @property
    def symm_relation(self):
        return symmetry_relation(self.initial_site_sym, self.final_site_sym)

    @property
    def same_config_from_init(self):
        return self.site_diff_from_initial.is_no_diff

    @property
    def defect_type(self):
        return judge_defect_type(self.site_diff)

    def __str__(self):
        sym_transition = f"{self.initial_site_sym} " \
                         f"-> {self.final_site_sym} ({self.symm_relation})"
        center_coords = pretty_coords(self.center)
        lines = [" -- defect structure info",
                 f"Defect type: {self.defect_type}",
                 f"Site symmetry: {sym_transition}",
                 f"Has same configuration from initial structure: "
                 f"{self.same_config_from_init}",
                 f"Drift distance: {self.drift_dist:5.3f}",
                 f"Defect center: {center_coords}"]

        def _site_info(header: str, site_info: List[SiteInfo]):
            if not site_info:
                return []
            _table = tabulate([[idx, elem, pretty_coords(coords)]
                              for idx, elem, coords in site_info],
                              tablefmt="plain")
            return [header, _table, ""]

        lines.extend(_site_info("Removed atoms:", self.site_diff.removed))
        lines.extend(_site_info("Added atoms:", self.site_diff.inserted))

        if self.same_config_from_init is False:
            lines.extend(_site_info("Removed atoms from initial structure:",
                                    self.site_diff_from_initial.removed))
            lines.extend(_site_info("Inserted atoms to initial structure:",
                                    self.site_diff_from_initial.inserted))

        min_d = min([d.distance_from_defect for d in self.displacements if d])
        cutoff = min_d * self.neighbor_cutoff_factor
        lines.append(f"Neighbor max distance {cutoff:5.3f}")

        lines.append("Displacements")
        idxs = [[i, d] for i, d in enumerate(self.displacements)
                if d is not None]
        table = [["Elem", "Dist", "Displace", "Angle", "Index",
                  "Initial site", "", "Final site", "Neighbor"]]
        print(self.neighbor_atom_indices)
        for final_idx, d in sorted(idxs,
                                   key=lambda y: y[1].distance_from_defect):
            is_neighbor = "T" if final_idx in self.neighbor_atom_indices else ""
            if d.distance_from_defect > defaults.show_structure_cutoff:
                break
            i_pos = pretty_coords(d.original_pos)
            f_pos = pretty_coords(d.final_pos)
            angle = int(round(d.angle, -1)) if d.angle else ""
            table.append([d.specie, round(d.distance_from_defect, 2),
                          round(d.displace_distance, 2), angle, final_idx,
                          i_pos, "->", f_pos, is_neighbor])
        lines.append(tabulate(table, tablefmt="plain"))

        return "\n".join(lines)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import os
from pathlib import Path

from monty.serialization import loadfn
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.grids import Grids
from pydefect.analyzer.refine_defect_structure import refine_defect_structure
from pydefect.cli.vasp.make_defect_charge_info import make_defect_charge_info
from pydefect.cli.vasp.get_defect_charge_state import get_defect_charge_state
from pydefect.input_maker.defect_entry import make_defect_entry
from pymatgen.core import Structure
from pymatgen.io.vasp import Chgcar
from vise.input_set.incar import ViseIncar
from vise.util.file_transfer import FileLink
from vise.util.logger import get_logger
from pymatgen.io.vasp.inputs import Poscar, Incar, Potcar

logger = get_logger(__name__)


def is_file(filename):
    return Path(filename).is_file() and os.stat(filename).st_size != 0


def calc_charge_state(args):
    poscar = Poscar.from_file(args.dir / "POSCAR")
    potcar = Potcar.from_file(args.dir / "POTCAR")
    incar = Incar.from_file(args.dir / "INCAR")
    charge_state = get_defect_charge_state(poscar, potcar, incar)
    logger.info(f"Charge state in {args.dir} is {charge_state}.")
    return charge_state


def make_defect_entry_main(args):
    charge_state = calc_charge_state(args)
    structure = Structure.from_file(args.dir / "POSCAR")
    defect_entry = make_defect_entry(name=args.name,
                                     charge=charge_state,
                                     perfect_structure=args.perfect,
                                     defect_structure=structure)
    defect_entry.to_json_file()


def make_parchg_dir(args):
    os.chdir(args.dir)
    if is_file("WAVECAR") is False:
        raise FileNotFoundError("WAVECAR does not exist or is empty.")

    try:
        calc_results: CalcResults = loadfn("calc_results.json")
    except FileNotFoundError:
        logger.info("Need to create calc_results.json beforehand.")
        raise
    calc_results.show_convergence_warning()

    # Increment index by 1 as VASP band index begins from 1.
    incar = ViseIncar.from_file("INCAR")
    band_edge_states = loadfn("band_edge_states.json")
    iband = [i + 1 for i in band_edge_states.band_indices_from_vbm_to_cbm]
    incar.update({"LPARD": True, "LSEPB": True, "KPAR": 1, "IBAND": iband})

    parchg = Path("parchg")
    parchg.mkdir()
    os.chdir("parchg")
    incar.write_file("INCAR")
    FileLink(Path("../WAVECAR")).transfer(Path.cwd())
    FileLink(Path("../POSCAR")).transfer(Path.cwd())
    FileLink(Path("../POTCAR")).transfer(Path.cwd())
    FileLink(Path("../KPOINTS")).transfer(Path.cwd())
    os.chdir("..")


def make_refine_defect_poscar(args):
    structure = refine_defect_structure(args.structure,
                                        args.defect_entry.anchor_atom_index,
                                        args.defect_entry.anchor_atom_coords)
    if structure:
        print(structure.to(fmt="poscar", filename=args.poscar_name))


def calc_grids(args):
    grids = Grids.from_chgcar(args.chgcar)
    grids.dump()


def make_defect_charge_info_main(args):
    band_idxs = [int(parchg.split(".")[-2]) - 1 for parchg in args.parchgs]
    parchgs = [Chgcar.from_file(parchg) for parchg in args.parchgs]
    defect_charge_info = make_defect_charge_info(
        parchgs, band_idxs, args.bin_interval, args.grids)
    defect_charge_info.to_json_file()
    plt = defect_charge_info.show_dist()
    plt.savefig("dist.pdf")





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from typing import Dict

from pydefect.analyzer.band_edge_states import BandEdgeStates
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_energy import DefectEnergy, DefectEnergyInfo
from pydefect.analyzer.unitcell import Unitcell
from pydefect.chem_pot_diag.chem_pot_diag import StandardEnergies
from pydefect.corrections.abstract_correction import Correction
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import IStructure


def make_defect_energy_info(defect_entry: DefectEntry,
                            calc_results: CalcResults,
                            correction: Correction,
                            perfect_calc_results: CalcResults,
                            standard_energies: StandardEnergies,
                            unitcell: Unitcell,
                            band_edge_states: BandEdgeStates = None
                            ) -> DefectEnergyInfo:
    atom_io = num_atom_differences(calc_results.structure,
                                   perfect_calc_results.structure)

    formation_energy = calc_results.energy - perfect_calc_results.energy
    formation_energy += defect_entry.charge * unitcell.vbm
    for k, v in atom_io.items():
        formation_energy -= standard_energies[k] * v

    is_shallow = band_edge_states.is_shallow if band_edge_states else None
    energy = DefectEnergy(formation_energy=formation_energy,
                          energy_corrections=correction.correction_dict,
                          is_shallow=is_shallow)

    return DefectEnergyInfo(defect_entry.name, defect_entry.charge,
                            atom_io=atom_io, defect_energy=energy)


def num_atom_differences(structure: IStructure,
                         ref_structure: IStructure,
                         ) -> Dict[str, int]:
    target_composition = structure.composition.as_dict()
    reference_composition = ref_structure.composition.as_dict()
    result = {}
    for k in set(target_composition.keys()) | set(reference_composition.keys()):
        n_atom_diff = int(target_composition[k] - reference_composition[k])
        if n_atom_diff:
            result[k] = n_atom_diff
    return result



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from itertools import cycle

from monty.design_patterns import singleton
from vise.defaults import DefaultsBase


@singleton
class Defaults(DefaultsBase):
    def __init__(self):
        self._symmetry_length_tolerance = 0.1
        self._symmetry_angle_tolerance = 5.0
        self._ewald_accuracy = 15.0
        self._e_above_hull = 1e-5
        self._cutoff_distance_factor = 1.3
        self._show_structure_cutoff = 5.0
        self._displace_distance = 0.2
        self._dist_tol = 1.0
        self._ele_neg_diff = 2.0
        self._similar_orb_criterion = 0.2
        self._similar_energy_criterion = 0.5  # in eV
        self._state_occupied_threshold = 0.20
        self._eigval_range = 1.0
        self._defect_energy_colors = \
            ["xkcd:blue", "xkcd:brown", "xkcd:crimson", "xkcd:darkgreen",
             "xkcd:gold", "xkcd:magenta", "xkcd:orange", "xkcd:darkblue",
             "xkcd:navy", "xkcd:red", "xkcd:olive", "xkcd:black", "xkcd:indigo"]
        self._abs_strange_energy = 100.0
        self._localized_orbital_radius = 3.0
        self._localized_orbital_fraction_wrt_uniform = 0.7

        self.set_user_settings(yaml_filename="pydefect.yaml")

    @property
    def symmetry_length_tolerance(self):
        return self._symmetry_length_tolerance

    @property
    def symmetry_angle_tolerance(self):
        return self._symmetry_angle_tolerance

    @property
    def ewald_accuracy(self):
        return self._ewald_accuracy

    @property
    def e_above_hull(self):
        return self._e_above_hull

    @property
    def cutoff_distance_factor(self):
        return self._cutoff_distance_factor

    @property
    def show_structure_cutoff(self):
        return self._show_structure_cutoff

    @property
    def displace_distance(self):
        return self._displace_distance

    @property
    def dist_tol(self):
        return self._dist_tol

    @property
    def ele_neg_diff(self):
        return self._ele_neg_diff

    @property
    def similar_orb_criterion(self):
        return self._similar_orb_criterion

    @property
    def similar_energy_criterion(self):
        return self._similar_energy_criterion

    @property
    def state_occupied_threshold(self):
        return self._state_occupied_threshold

    @property
    def state_unoccupied_threshold(self):
        return 1 - self._state_occupied_threshold

    @property
    def eigval_range(self):
        return self._eigval_range

    @property
    def defect_energy_colors(self):
        return cycle(self._defect_energy_colors)

    @property
    def abs_strange_energy(self):
        return self._abs_strange_energy

    @property
    def localized_orbital_radius(self):
        return self._localized_orbital_radius

    @property
    def localized_orbital_fraction_wrt_uniform(self):
        return self._localized_orbital_fraction_wrt_uniform


defaults = Defaults()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import argparse
import sys
from dataclasses import dataclass
from typing import Dict, List

import numpy as np
from crystal_toolkit.core.scene import Scene, Surface
from monty.json import MSONable
from monty.serialization import loadfn
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import Structure
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Chgcar, VolumetricData
from skimage.measure import marching_cubes
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class SceneDicts(MSONable, ToJsonFileMixIn):
    scene_dicts: Dict[str, Dict[str, np.ndarray]]
    structure_graph: StructureGraph

    @property
    def scenes(self):
        result = {}
        if len(self.scene_dicts) == 0:
            return {"empty_scene": Scene("empty", contents=[])}
        for k, v in self.scene_dicts.items():
            pos = [vert for triangle in v["vertices"][v["faces"]].tolist()
                   for vert in triangle]
            result[k] = Scene(k, contents=[Surface(positions=pos)])
        return result

    def as_dict(self) -> dict:
        """
        A JSON serializable dict representation of an object.
        """
        scene_dicts = {k: {k2: v2.tolist() for k2, v2 in v.items()}
                       for k, v in self.scene_dicts.items()}

        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "scene_dicts": scene_dicts,
                "structure_graph": self.structure_graph.as_dict()}

    @classmethod
    def from_dict(cls, d):
        scene_dicts = {k: {k2: np.array(v2) for k2, v2 in v.items()}
                       for k, v in d["scene_dicts"].items()}
        return cls(scene_dicts, StructureGraph.from_dict(d["structure_graph"]))


def make_scene_dicts(parchg_list, defect_pos, level=None):
    band_indices = [c.split(".")[-2] for c in parchg_list]
    # parchgs = [Chgcar.from_file(c) for c in parchg_list]
    parchgs = []
    for c in parchg_list:
        try:
            parchgs.append(Chgcar.from_file(c))
        except ValueError:
            continue
    if len(parchgs) == 0:
        return None
    structure: Structure = parchgs[0].structure
    if len(parchg_list) > 1:
        for parchg in parchgs[1:]:
            assert structure == parchg.structure

    lattice_matrix = parchgs[0].structure.lattice.matrix

    # centering to [0.5, 0.5, 0.5]
    shift_vector = np.array([0.5 - pos for pos in defect_pos])
    shape = parchgs[0].data["total"].shape
    shift_numbers = np.array(np.rint(shape * shift_vector), dtype=np.int)
    actual_shift_vector = shift_numbers / shape

    step_size = int(min(parchgs[0].dim) / 30)
    results = {}
    for name, parchg in zip(band_indices, parchgs):
        for spin in [Spin.up, Spin.down]:
            spin_str = "up" if spin is Spin.up else "down"
            key = name + "_" + spin_str
            data = parchg.spin_data[spin]
            x_shifted = np.roll(data, shift_numbers[0], axis=0)
            xy_shifted = np.roll(x_shifted, shift_numbers[1], axis=1)
            xyz_shifted = np.roll(xy_shifted, shift_numbers[2], axis=2)
            try:
                vertices, faces = get_vertices_and_faces(xyz_shifted,
                                                         lattice_matrix,
                                                         step_size=step_size,
                                                         level=level)
                results[key] = {"vertices": vertices, "faces": faces}
            except RuntimeError:
                continue

    structure.translate_sites(indices=list(range(len(structure))),
                              vector=actual_shift_vector)

    graph = StructureGraph.with_empty_graph(structure=structure)
    return SceneDicts(results, structure_graph=graph)


def get_vertices_and_faces(array_data, lattice_matrix, step_size=10, level=None):
    level = level or 0.5
    # normals and values are removed.
    vertices, faces, _, _ = marching_cubes(array_data,
                                           level=level * array_data.max(),
                                           step_size=step_size)
    # normalize to fractional coordinates
    vertices = vertices / array_data.shape
    # transform to cartesian
    vertices = np.dot(vertices - 0.5, lattice_matrix)
    return vertices, faces


@dataclass
class SingleScene(MSONable):
    scene_dict: Dict[str, np.ndarray]

    def scene(self, name=""):
        faces = self.scene_dict["faces"]
        pos = [vert for triangle in self.scene_dict["vertices"][faces].tolist()
               for vert in triangle]
        return Scene(name=name, contents=[Surface(positions=pos)])

    def as_dict(self) -> dict:
        scene_dict = {k: v.tolist() for k, v in self.scene_dict.items()}
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "scene_dict": scene_dict}

    @classmethod
    def from_dict(cls, d):
        scene_dicts = {k: np.array(v) for k, v in d["scene_dict"].items()}
        return cls(scene_dicts)


def make_single_scene(volmetric_data: VolumetricData, step_size=3, level=None):
    lattice_matrix = volmetric_data.structure.lattice.matrix
    data = volmetric_data.data["total"]
    vertices, faces = get_vertices_and_faces(data,
                                             lattice_matrix,
                                             step_size=step_size,
                                             level=level)
    return SingleScene({"vertices": vertices, "faces": faces})


def make_scene_dicts_json(parchg_list: List[str],
                          calc_results: CalcResults,
                          perfect_calc_results: CalcResults,
                          scene_level):
    structure_analyzer = DefectStructureComparator(calc_results.structure,
                                                   perfect_calc_results.structure)
    defect_pos = structure_analyzer.defect_center_coord
    scene_dicts = make_scene_dicts(parchg_list, defect_pos, level=scene_level)
    scene_dicts.to_json_file()


def parse_args(args):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("--parchgs", type=str, nargs="+")
    parser.add_argument("-c", "--calc_results", type=loadfn)
    parser.add_argument("-p", "--perfect_calc_results", type=loadfn)
    parser.add_argument("-l", "--level", type=float)
    return parser.parse_args(args)


if __name__ == "__main__":
    x = parse_args(sys.argv[1:])
    make_scene_dicts_json(x.parchgs, x.calc_results, x.perfect_calc_results,
                          x.level)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
from numpy import dot, cross
from numpy.linalg import norm

from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection, PotentialSite
from pydefect.corrections.ewald import Ewald
from pydefect.defaults import defaults
from pydefect.util.error_classes import SupercellError
from vise.util.logger import get_logger


logger = get_logger(__name__)


def make_efnv_correction(charge: int,
                         calc_results: CalcResults,
                         perfect_calc_results: CalcResults,
                         dielectric_tensor: np.array,
                         accuracy: float = defaults.ewald_accuracy,
                         unit_conversion: float = 180.95128169876497):
    """
    Notes:
    (1) The formula written in YK2014 need to be divided by 4pi in the SI unit.
    (2) When assuming an element charge locate at the defect_coords and
        angstrom for length, relative dielectric tensor, Multiply
        elementary_charge * 1e10 / epsilon_0 = 180.95128169876497
        to make potential in V.
    """
    if calc_results.structure.lattice != perfect_calc_results.structure.lattice:
        raise SupercellError("The lattice constants for defect and perfect "
                             "models are different")

    structure_analyzer = DefectStructureComparator(
        calc_results.structure, perfect_calc_results.structure)
    defect_coords = structure_analyzer.defect_center_coord
    lattice = calc_results.structure.lattice
    ewald = Ewald(lattice.matrix, dielectric_tensor, accuracy=accuracy)
    point_charge_correction = \
        0.0 if not charge else - ewald.lattice_energy * charge ** 2

    defect_region_radius = calc_max_sphere_radius(lattice.matrix)

    sites = []
    for d, p in structure_analyzer.atom_mapping.items():
        specie = str(calc_results.structure[d].specie)
        frac_coords = calc_results.structure[d].frac_coords
        distance, _ = lattice.get_distance_and_image(defect_coords, frac_coords)
        pot = calc_results.potentials[d] - perfect_calc_results.potentials[p]

        coord = calc_results.structure[d].frac_coords
        rel_coord = [x - y for x, y in zip(coord, defect_coords)]
        if distance <= defect_region_radius:
            pc_potential = None
        else:
            if charge == 0:
                pc_potential = 0
            else:
                pc_potential = ewald.atomic_site_potential(rel_coord) * charge
                pc_potential *= unit_conversion

        sites.append(PotentialSite(specie, distance, pot, pc_potential))

    return ExtendedFnvCorrection(
        charge=charge,
        point_charge_correction=point_charge_correction * unit_conversion,
        defect_region_radius=defect_region_radius,
        sites=sites,
        defect_coords=tuple(defect_coords))


def calc_max_sphere_radius(lattice_matrix) -> float:
    """Calculate Maximum radius of a sphere fitting inside the unit cell.

    Calculate three distances between two parallel planes using
    (a_i x a_j) . a_k / |a_i . a_j| """
    distances = np.zeros(3, dtype=float)
    for i in range(3):
        a_i_a_j = cross(lattice_matrix[i - 2], lattice_matrix[i - 1])
        a_k = lattice_matrix[i]
        distances[i] = abs(dot(a_i_a_j, a_k)) / norm(a_i_a_j)
    return max(distances) / 2.0



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
import pytest
from numpy import pi, sqrt, exp, cos
from pydefect.corrections.ewald import Ewald
from scipy.special import erfc
from vise.tests.helpers.assertion import assert_msonable


@pytest.fixture
def ewald():
    return Ewald(lattice=np.array([[2, 0, 0], [0, 2, 0], [0, 0, 2]]),
                 dielectric_tensor=np.array([[3, 0, 0], [0, 3, 0], [0, 0, 3]]),
                 ewald_param=0.1,
                 accuracy=0.1)


def test_msonable(ewald):
    assert_msonable(ewald)


def test_ewald(ewald):
    np.testing.assert_almost_equal(
        ewald.lattice, np.array([[2, 0, 0], [0, 2, 0], [0, 0, 2]]))
    np.testing.assert_almost_equal( # 0.5 * 2 * pi
        ewald.rec_lattice, np.array([[pi, 0, 0], [0, pi, 0], [0, 0, pi]]))
    np.testing.assert_almost_equal(ewald.volume, 8.0)
    np.testing.assert_almost_equal(ewald.cube_root_vol, 2.0)
    np.testing.assert_almost_equal(ewald.det_epsilon, 27)
    np.testing.assert_almost_equal(ewald.root_epsilon, 3 * np.sqrt(3))
    np.testing.assert_almost_equal(
        ewald.epsilon_inv, np.array([[1/3, 0, 0], [0, 1/3, 0], [0, 0, 1/3]]))
    np.testing.assert_almost_equal(
        ewald.mod_ewald_param, 0.1 / 2.0 * 3 * np.sqrt(3))
    np.testing.assert_almost_equal(
        ewald.diff_pot, -0.25 / 8.0 / (0.1 / 2.0 * 3 * np.sqrt(3)) ** 2)


def test_ewald_vector_nums():
    ewald = Ewald(lattice=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                  dielectric_tensor=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                  ewald_param=1,
                  accuracy=1.0001)
    assert ewald.r_vector_nums == [2, 2, 2]

    ewald = Ewald(lattice=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                  dielectric_tensor=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                  ewald_param=1,
                  accuracy=6.28318531 / 2 + 0.0001)
    assert ewald.g_vector_nums == [2, 2, 2]


def test_ewald_xyz(ewald):
    actual = ewald.xyz([1, 1, 1], shift=[0.1, 0.1, 0.1])
    assert len(actual) == 27
    np.testing.assert_almost_equal(actual[0], np.array([-1.1, -1.1, -1.1]))


def test_ewald_lattice_set(ewald):
    np.testing.assert_almost_equal(ewald.r_lattice_set()[0], np.array([-2, -2, -2]))
    np.testing.assert_almost_equal(ewald.g_lattice_set()[0], np.array([-3.14159265, -3.14159265, -3.14159265]))


def test_ewald_real(ewald, mocker):
    mock = mocker.patch("pydefect.corrections.ewald.Ewald.r_lattice_set")
    mock.return_value = np.array([[-2, -2, -2]])
    root_r_inv_epsilon_r = sqrt(2 / 3 * 2 * 3)
    actual = ewald.ewald_real(include_self=False, shift=[0, 0, 0])
    expected = erfc(ewald.mod_ewald_param * root_r_inv_epsilon_r) / root_r_inv_epsilon_r / (4 * pi * ewald.root_epsilon)
    assert actual == expected


def test_ewald_rec(ewald, mocker):
    mock = mocker.patch("pydefect.corrections.ewald.Ewald.g_lattice_set")
    mock.return_value = np.array([[2, 2, 2]])
    g_epsilon_g = 2 * 3 * 2 * 3
    actual = ewald.ewald_rec([0.2, 0.2, 0.2])
    cos_term = cos(2 * 2 * 0.2 * 3)
    expected = exp(-g_epsilon_g / 4 / ewald.mod_ewald_param ** 2) / g_epsilon_g * cos_term / ewald.volume
    assert actual == expected

#
# def test_ewald_speed():
#     ewald = Ewald(lattice=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
#                   dielectric_tensor=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
#                   accuracy=30)
#     assert ewald.ewald_real(include_self=False, shift=None) == 0.11400074023354886



"""
TODO
- Create a container Ewald class
- 

DONE
"""


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pydefect.analyzer.defect_energy import DefectEnergy, DefectEnergies, \
    DefectEnergySummary
from pydefect.analyzer.defect_energy_plotter import DefectEnergyMplPlotter, \
    DefectEnergiesMplSettings, DefectEnergyPlotlyPlotter
from pydefect.defaults import defaults
from vise.util.dash_helper import show_png

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


def test_defect_energies_mpl_settings():
    mpl_defaults = DefectEnergiesMplSettings()
    assert next(mpl_defaults.colors) == next(defaults.defect_energy_colors)

    mpl_defaults = DefectEnergiesMplSettings(colors=["black"])
    assert next(mpl_defaults.colors) == "black"


@pytest.fixture
def defect_energy_summary():
    de1 = DefectEnergy(0.0, {"PC correction": 2.0}, False)
    de2 = DefectEnergy(-5.0, {"PC correction": 3.0}, True)

    defect_energies = {"Va_Mg1": DefectEnergies(atom_io={"Mg": -1},
                                                charges=[0, 1],
                                                defect_energies=[de1, de2])}
    return DefectEnergySummary(title="MgAl2O4",
                               defect_energies=defect_energies,
                               rel_chem_pots={"A": {"Mg": -1}},
                               cbm=7.0,
                               supercell_vbm=-1.0,
                               supercell_cbm=4.0)


def test_defect_energies_mpl_plot(defect_energy_summary):
    plotter = DefectEnergyMplPlotter(
        defect_energy_summary=defect_energy_summary,
        chem_pot_label="A",
        allow_shallow=True,
        with_corrections=True)
    plotter.construct_plot()
    plotter.plt.show()


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="psutil does not exist")
def test_defect_energies_plotly_plot(defect_energy_summary):
    plotter = DefectEnergyPlotlyPlotter(
        defect_energy_summary=defect_energy_summary,
        chem_pot_label="A",
        allow_shallow=False,
        with_corrections=True)
    fig = plotter.create_figure()
    show_png(fig)


"""
TODO
- Show all energies with thinner lines.
- show_transition_levels

- Add labels with optimize positions
https://github.com/cphyc/matplotlib-label-lines
https://github.com/Phlya/adjustText

- Add boundary points with open circles for shallow defects.
DONE
"""


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.unitcell import Unitcell
from pydefect.cli.main_util import parse_args_main_util
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection
from pydefect.defaults import defaults


def test_make_cpd_options():
    parsed_args = parse_args_main_util(["cefm",
                                        "-e", "Mg", "O",
                                        "-a", "atom_energies.yaml"])
    expected = Namespace(
        elements=["Mg", "O"],
        atom_energy_yaml="atom_energies.yaml",
        func=parsed_args.func)
    assert parsed_args == expected


def test_add_interstitials_from_local_extrema(mocker):

    mock_loadfn_main = mocker.patch("pydefect.cli.main.loadfn")
    mock_loadfn_main_util = mocker.patch("pydefect.cli.main_util.loadfn")

    parsed_args = parse_args_main_util(["ai",
                                        "-s", "supercell_info.json",
                                        "--local_extrema", "local_extrema.json",
                                        "--indices", "1", "2"])
    expected = Namespace(
        supercell_info=mock_loadfn_main.return_value,
        local_extrema=mock_loadfn_main_util.return_value,
        indices=[1, 2],
        func=parsed_args.func)
    assert parsed_args == expected
    mock_loadfn_main.assert_called_once_with("supercell_info.json")
    mock_loadfn_main_util.assert_called_once_with("local_extrema.json")


def test_defect_vesta_file_wo_options():
    parsed_args = parse_args_main_util(["dvf", "-d", "Va_O1_0", "Va_O1_1"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        cutoff=defaults.show_structure_cutoff,
        min_displace_w_arrows=0.1,
        arrow_factor=3.0,
        title=None,
        func=parsed_args.func)
    assert parsed_args == expected


def test_defect_vesta_file_w_options():
    parsed_args = parse_args_main_util(["dvf",
                                        "-d", "Va_O1_0", "Va_O1_1",
                                        "--cutoff", "1.0",
                                        "--min_displace_w_arrows", "2.0",
                                        "--arrow_factor", "10.0",
                                        "--title", "title"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        cutoff=1.0,
        min_displace_w_arrows=2.0,
        arrow_factor=10.0,
        title="title",
        func=parsed_args.func)
    assert parsed_args == expected


def test_gkfo_correction(mocker):
    mock_i_correction = mocker.Mock(spec=ExtendedFnvCorrection, autospec=True)
    mock_i_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_f_calc_results = mocker.Mock(spec=CalcResults, autospec=True)

    def side_effect(filename):
        if filename == "a/correction.json":
            return mock_i_correction
        elif filename == "a/calc_results.json":
            return mock_i_calc_results
        elif filename == "a/absorption/calc_results.json":
            return mock_f_calc_results
        else:
            raise ValueError

    mocker.patch("pydefect.cli.main_util.loadfn", side_effect=side_effect)
    mock_unitcell = mocker.patch("pydefect.cli.main.Unitcell")

    parsed_args = parse_args_main_util([
        "gkfo",
        "-iefnv", "a/correction.json",
        "-cd", "1",
        "-icr", "a/calc_results.json",
        "-fcr", "a/absorption/calc_results.json",
        "-u", "unitcell.json"])

    expected = Namespace(
        initial_efnv_correction=mock_i_correction,
        initial_calc_results=mock_i_calc_results,
        final_calc_results=mock_f_calc_results,
        charge_diff=1,
        unitcell=mock_unitcell.from_yaml.return_value,
        func=parsed_args.func)

    assert parsed_args == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from dataclasses import dataclass
from typing import List, Optional

from monty.json import MSONable
from pydefect.input_maker.append_interstitial import append_interstitial
from pydefect.util.coords import pretty_coords
from pydefect.util.structure_tools import Coordination
from pymatgen.core import Structure
from tabulate import tabulate
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.typing import Coords


@dataclass
class CoordInfo(MSONable):
    site_symmetry: str
    coordination: Coordination
    frac_coords: List[Coords]
    quantities: List[float] = None


@dataclass
class VolumetricDataAnalyzeParams(MSONable):
    threshold_frac: Optional[float]
    threshold_abs: Optional[float]
    min_dist: float
    tol: float
    radius: float


@dataclass
class VolumetricDataLocalExtrema(MSONable, ToJsonFileMixIn):
    unit_cell: Structure
    is_min: bool
    extrema_points: List[CoordInfo]
    info: str
    params: VolumetricDataAnalyzeParams

    def __str__(self):
        min_or_max = "min" if self.is_min else "max"
        extrema_points = [["#", "site_sym", "coordination", "frac_coords", "quantity"]]
        for idx, ep in enumerate(self.extrema_points, 1):
            extrema_points.append([idx,
                                   ep.site_symmetry,
                                   ep.coordination.distance_dict,
                                   pretty_coords(ep.frac_coords[0]),
                                   f"{ep.quantities[0]:.2g}"])
        lines = [f"info: {self.info}",
                 f"min_or_max: {min_or_max}",
                 f"extrema_points:",
                 tabulate(extrema_points, tablefmt="plain")]
        return "\n".join(lines)

    def append_sites_to_supercell_info(self, supercell_info, indices):
        frac_coords, infos = [], []
        for idx, ep in enumerate(self.extrema_points, 1):
            if idx in indices:
                frac_coords.append(ep.frac_coords[0])
                infos.append(f"{self.info} #{idx}")

        result = append_interstitial(supercell_info,
                                     self.unit_cell,
                                     frac_coords=frac_coords,
                                     infos=infos)
        return result




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy
from pathlib import Path

import pytest
from pydefect.analyzer.band_edge_states import BandEdgeEigenvalues, \
    BandEdgeStates, OrbitalInfo, BandEdgeOrbitalInfos, PerfectBandEdgeState, \
    EdgeInfo, BandEdgeState, LocalizedOrbital, pretty_orbital
from pydefect.defaults import defaults
from vise.tests.helpers.assertion import assert_msonable, assert_json_roundtrip


@pytest.fixture
def band_edge_eigenvalues():
    return BandEdgeEigenvalues(energies_and_occupations=[[[[0.0, 1.0]]]],
                               kpt_coords=[[0.0, 0.0, 0.0]],
                               lowest_band_index=10)


def test_band_edge_eigenvalues_msonable(band_edge_eigenvalues):
    assert_msonable(band_edge_eigenvalues)


def test_band_edge_eigenvalues_to_json_file(band_edge_eigenvalues, tmpdir):
    assert_json_roundtrip(band_edge_eigenvalues, tmpdir)


@pytest.fixture
def orbital_info():
    return OrbitalInfo(energy=1.0,
                       orbitals={"Mn": [0.5, 0.4, 0.0, 0.0]},
                       occupation=1.0,
                       participation_ratio=0.1)


def test_orbital_info_msonable(orbital_info):
    assert_msonable(orbital_info)


def test_pretty_orbital(orbital_info):
    assert pretty_orbital(orbital_info.orbitals) == "Mn-s: 0.50, Mn-p: 0.40"


@pytest.fixture
def band_edge_orbital_infos(orbital_info):
    return BandEdgeOrbitalInfos(orbital_infos=[[[orbital_info]]],
                                kpt_coords=[(0.0, 0.0, 0.0)],
                                kpt_weights=[1.0],
                                lowest_band_index=10,
                                fermi_level=0.5)


def test_kpt_idx(band_edge_orbital_infos):
    assert band_edge_orbital_infos.kpt_idx([0.0, 0.0, 0.0]) == 0


def test_band_edge_orbital_info_json_roundtrip(band_edge_orbital_infos, tmpdir):
    assert_json_roundtrip(band_edge_orbital_infos, tmpdir)


def test_band_edge_orbital_info_energies_occupations(band_edge_orbital_infos):
    actual = band_edge_orbital_infos.energies_and_occupations
    expected = [[[[1.0, 1.0]]]]
    assert actual == expected


def test_band_edge_orbital_info_repr():
    orbital_info_1 = OrbitalInfo(energy=1.23456789,
                                 orbitals={"Mn": [0.5, 0.4, 0.0, 0.0],
                                           "Fe": [0.0, 0.0, 0.0, 0.0]},
                                 occupation=0.99998,
                                 participation_ratio=0.1111111)
    orbital_info_2 = OrbitalInfo(energy=2.34567890,
                                 orbitals={"Mn": [0.1, 0.1, 0.0, 0.0],
                                           "Fe": [0.0, 0.0, 0.0, 0.0]},
                                 occupation=0.0001,
                                 participation_ratio=0.2222222)

    band_edge_orbital_infos =  \
        BandEdgeOrbitalInfos(orbital_infos=[[[orbital_info_1, orbital_info_2]]],
                             kpt_coords=[(0.0, 0.0, 0.0)],
                             kpt_weights=[1.0],
                             lowest_band_index=10,
                             fermi_level=0.5)

    actual = band_edge_orbital_infos.__str__()
    print(actual)
    expected = """ -- band-edge orbitals info
K-points info
Index  Coords                    Weight
1      ( 0.000,  0.000,  0.000)  1.000

Band info near band edges
Index  Kpoint index  Energy  Occupation  P-ratio  Orbital
11     1             1.23    1.0         0.1      Mn-s: 0.50, Mn-p: 0.40
--
12     1             2.35    0.0         0.2
--
"""
    assert actual == expected


@pytest.fixture
def perfect_band_edge_state(orbital_info):
    vbm_info = EdgeInfo(band_idx=10, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=orbital_info)
    return PerfectBandEdgeState(vbm_info=vbm_info, cbm_info=vbm_info)


def test_perfect_edge_state_json_roundtrip(perfect_band_edge_state, tmpdir):
    assert_json_roundtrip(perfect_band_edge_state, tmpdir)


@pytest.fixture
def band_edge_states(orbital_info):
    vbm_info = EdgeInfo(band_idx=10, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=orbital_info)
    cbm_info = EdgeInfo(band_idx=12, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=orbital_info)
    localized_orbital = LocalizedOrbital(
        band_idx=11, ave_energy=1.9, occupation=1.0,
        orbitals={"H": [1.0, 0.0, 0.0, 0.0]})
    band_edge_state = BandEdgeState(vbm_info=vbm_info, cbm_info=cbm_info,
                                    vbm_orbital_diff=0.5,
                                    cbm_orbital_diff=0.5,
                                    localized_orbitals=[localized_orbital],
                                    vbm_hole_occupation=defaults.state_occupied_threshold - 1e-5,
                                    cbm_electron_occupation=defaults.state_occupied_threshold + 1e-5,
                                    )
    return BandEdgeStates(states=[band_edge_state])


def test_band_edge_states_json_roundtrip(band_edge_states, tmpdir):
    assert_json_roundtrip(band_edge_states, tmpdir)


def test_band_edge_states_band_indices(band_edge_states):
    assert band_edge_states.band_indices_from_vbm_to_cbm == [10, 11, 12]


def test_band_edge_states_has_donor_phs(band_edge_states):
    assert band_edge_states.has_donor_phs is True
    assert band_edge_states.has_acceptor_phs is False
    assert band_edge_states.has_unoccupied_localized_state is False
    assert band_edge_states.has_occupied_localized_state is True


def test_band_edge_states_str(band_edge_states):
    expected = f""" -- band-edge states info
Spin-up
     Index  Energy  P-ratio  Occupation  OrbDiff  Orbitals                K-point coords
VBM  11     1.000   0.10     1.00        0.50     Mn-s: 0.50, Mn-p: 0.40  ( 0.000,  0.000,  0.000)
CBM  13     1.000   0.10     1.00        0.50     Mn-s: 0.50, Mn-p: 0.40  ( 0.000,  0.000,  0.000)
vbm has acceptor phs: False ({defaults.state_occupied_threshold:.3f} vs. {defaults.state_occupied_threshold})
cbm has donor phs: True ({defaults.state_occupied_threshold:.3f} vs. {defaults.state_occupied_threshold})
---
Localized Orbital(s)
Index  Energy  P-ratio  Occupation  Orbitals
12     1.900   None     1.00        H-s: 1.00
"""
    print(band_edge_states)
    assert band_edge_states.__str__() == expected


def test_band_edge_states_str2(band_edge_states):
    copy = deepcopy(band_edge_states)
    copy.states[0].localized_orbitals = [LocalizedOrbital(
        band_idx=11, ave_energy=1.9, occupation=1.0,
        orbitals={"H": [1.0, 0.0, 0.0, 0.0]}, radius=1.000001,
        center=[0.1234, 0.0, 0.0])]
    actual = copy.__str__()
    print(actual)
    expected = f""" -- band-edge states info
Spin-up
     Index  Energy  P-ratio  Occupation  OrbDiff  Orbitals                K-point coords
VBM  11     1.000   0.10     1.00        0.50     Mn-s: 0.50, Mn-p: 0.40  ( 0.000,  0.000,  0.000)
CBM  13     1.000   0.10     1.00        0.50     Mn-s: 0.50, Mn-p: 0.40  ( 0.000,  0.000,  0.000)
vbm has acceptor phs: False ({defaults.state_occupied_threshold:.3f} vs. {defaults.state_occupied_threshold})
cbm has donor phs: True ({defaults.state_occupied_threshold:.3f} vs. {defaults.state_occupied_threshold})
---
Localized Orbital(s)
Index  Energy  P-ratio  Occupation  Orbitals   Radius  Center
12     1.900   None     1.00        H-s: 1.00  1.00    ( 0.123,  0.000,  0.000)
"""
    assert actual == expected


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest
from pydefect.input_maker.defect_entry import DefectEntry, make_defect_entry, \
    PerturbedSite
from pymatgen.core import Lattice, IStructure

# "H" at [0.0, 0.0, 0.0] is removed here.
from vise.tests.helpers.assertion import assert_msonable, assert_json_roundtrip

perf_coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5],
               [0.0, 0.0, 0.5], [0.0, 0.5, 0.0], [0.5, 0.0, 0.0], [0.5, 0.5, 0.5]]

coords = deepcopy(perf_coords)
coords.pop(0)

perturbed_coords = \
    [[0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5],
     [0.0, 0.0, 0.501], [0.0, 0.501, 0.0], [0.501, 0.0, 0.0], [0.5, 0.5, 0.5]]

perfect = IStructure(
    Lattice.cubic(10.0), ["H"] * 4 + ["He"] * 4, perf_coords)
defect_structure = IStructure(
    Lattice.cubic(10.0), ["H"] * 3 + ["He"] * 4, coords)
perturbed_defect = IStructure(
    Lattice.cubic(10.0), ["H"] * 3 + ["He"] * 4, perturbed_coords)


@pytest.fixture
def defect_entry():
    perturbed_sites = (
        PerturbedSite("He", 5.0, (0.0, 0.0, 0.5), (0.0, 0.0, 0.501), 0.1),
        PerturbedSite("He", 5.0, (0.0, 0.5, 0.0), (0.0, 0.501, 0.0), 0.1),
        PerturbedSite("He", 5.0, (0.5, 0.0, 0.0), (0.501, 0.0, 0.0), 0.1))
    return DefectEntry(name="Va_O1",
                       charge=1,
                       structure=defect_structure,
                       site_symmetry="m-3m",
                       defect_center=(0.0, 0.0, 0.0),
                       perturbed_structure=perturbed_defect,
                       perturbed_sites=perturbed_sites,
                       perturbed_site_symmetry="1")


def test_hashable(defect_entry):
    d = {defect_entry: 1}


def test_perturbed_site_indices(defect_entry):
    assert defect_entry.perturbed_site_indices == [3, 4, 5]


def test_full_name(defect_entry):
    assert defect_entry.full_name == "Va_O1_1"


def test_anchor_atom_index(defect_entry):
    assert defect_entry.anchor_atom_index == 6


def test_anchor_atom_coords(defect_entry):
    np.testing.assert_array_almost_equal(defect_entry.anchor_atom_coords,
                                         np.array([0.5, 0.5, 0.5]))


def test_msonable(defect_entry):
    assert_msonable(defect_entry)


def test_json_roundtrip(defect_entry, tmpdir):
    assert_json_roundtrip(defect_entry, tmpdir)


def test_defect_entry(defect_entry, tmpdir):
    tmpdir.chdir()
    defect_entry.to_prior_info(filename="a.yaml")
    expected = """charge: 1
"""
    assert Path("a.yaml").read_text() == expected


def test_defect_entry_str(defect_entry, tmpdir):
    actual = defect_entry.__str__()
    expected = """ -- defect entry info
name: Va_O1_1
site symmetry: m-3m
defect center: ( 0.000,  0.000,  0.000)
perturbed sites:
elem dist   initial_coords             perturbed_coords         displacement
  He 5.00 ( 0.000,  0.000,  0.500) -> ( 0.000,  0.000,  0.501)    0.10
  He 5.00 ( 0.000,  0.500,  0.000) -> ( 0.000,  0.501,  0.000)    0.10
  He 5.00 ( 0.500,  0.000,  0.000) -> ( 0.501,  0.000,  0.000)    0.10"""
    assert actual == expected


def test_make_defect_entry(defect_entry):
    relaxed_coords = \
        [[0.25, 0.0, 0.0], [0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5],
         [0.0, 0.0, 0.51], [0.0, 0.51, 0.0], [0.5, 0.0, 0.0], [0.5, 0.5, 0.5]]
    relaxed_defect = IStructure(
        Lattice.cubic(10.0), ["Li"] + ["H"] * 3 + ["He"] * 4, relaxed_coords)

    unrelaxed_coords = \
        [[0.25, 0.0, 0.0], [0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5],
         [0.0, 0.0, 0.5], [0.0, 0.5, 0.0], [0.5, 0.0, 0.0], [0.5, 0.5, 0.5]]

    unrelaxed_defect = IStructure(
        Lattice.cubic(10.0), ["Li"] + ["H"] * 3 + ["He"] * 4, unrelaxed_coords)

    actual = make_defect_entry(name="Va_O1",
                               charge=1,
                               perfect_structure=perfect,
                               defect_structure=relaxed_defect)

    expected = DefectEntry(name="Va_O1",
                           charge=1,
                           structure=unrelaxed_defect,
                           perturbed_structure=None,
                           site_symmetry="4mm",
                           defect_center=(0.125, 0.0, 0.0))

    assert actual == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from itertools import groupby
from pathlib import Path
from typing import List, Optional, Dict

from monty.serialization import loadfn
from pydefect.chem_pot_diag.chem_pot_diag import CompositionEnergy, \
    CompositionEnergies
from pydefect.util.mp_tools import MpQuery, MpEntries
from pymatgen.core import Composition
from pymatgen.entries.computed_entries import ComputedEntry
from vise.atom_energies.atom_energy import mp_energies
from vise.util.logger import get_logger

parent = Path(__file__).parent
logger = get_logger(__name__)


def make_composition_energies_from_mp(elements: List[str],
                                      atom_energy_yaml: Optional[str] = None,
                                      ) -> CompositionEnergies:
    """Obtain the energies from Materials Project.

    When the atom_energy_yaml is provided, the total energies are aligned
    via atom energies.
    """
    entries: List[ComputedEntry] = MpEntries(elements).materials
    comp_es = {}
    if atom_energy_yaml:
        energies = loadfn(atom_energy_yaml)
        diff = {e: energies[e] - mp_energies[e] for e in elements}
    else:
        diff = {e: 0.0 for e in elements}

    for e in entries:
        key = e.composition
        energy = e.energy
        for k, v in key.as_dict().items():
            energy += diff[k] * v
        comp_es[key] = CompositionEnergy(energy, e.entry_id)
    comp_es = remove_higher_energy_comp(comp_es)
    return CompositionEnergies(comp_es)


def remove_higher_energy_comp(
        comp_energies: Dict[Composition, CompositionEnergy]):
    comp_energy_pairs = [[k, v] for k, v in comp_energies.items()]
    result = {}
    key = lambda x: Composition(x[0]).reduced_formula
    sorted_comp_energy_pairs = sorted(comp_energy_pairs, key=key)
    for _, grouped_k_v in groupby(sorted_comp_energy_pairs, key=key):
        formula, comp_e = min(list(grouped_k_v),
                              key=lambda y: (y[1].energy
                                             / Composition(y[0]).num_atoms))
        result[formula] = comp_e
    return result


if __name__ == "__main__":
    print(make_composition_energies_from_mp(["Mg", "O"]))


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from copy import deepcopy

import pytest
from pydefect.input_maker.append_interstitial import append_interstitial
from pydefect.input_maker.supercell_info import Interstitial
from pydefect.util.error_classes import NotPrimitiveError
from pymatgen.core import Structure, Lattice


def test_add_interstitial(cubic_supercell_info_wo_int):
    primitive = Structure(Lattice.rhombohedral(7.071068, 60),
                          species=["H", "He"],
                          coords=[[0.0]*3, [0.5]*3])
    new_supercell_info = append_interstitial(cubic_supercell_info_wo_int,
                                             primitive,
                                             [[1/4, 1/4, 1/4]],
                                             infos=["test1"])
    expected = Interstitial(frac_coords=[1/8, 1/8, 1/8],
                            site_symmetry="-43m",
                            info="test1")
    assert new_supercell_info.interstitials[0] == expected


def test_add_interstitial2(mocker, simple_cubic):
    mock_supercell_info = mocker.Mock()
    mock_supercell_info.unitcell_structure = simple_cubic
    mock_supercell_info.transformation_matrix = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    mock_supercell_info.interstitials = []

    unitcell = Structure.from_dict(simple_cubic.as_dict())
    new_supercell_info = append_interstitial(mock_supercell_info,
                                             unitcell,
                                             [(0.0, 0.0, 0.1)],
                                             infos=["test1"])
    expected = Interstitial(frac_coords=[0.0, 0.0, 0.05],
                            site_symmetry="4mm",
                            info="test1")
    assert new_supercell_info.interstitials[0] == expected


def test_add_interstitial_not_primitive_error(cubic_supercell_info_wo_int):
    conventional_cell = Structure(Lattice.cubic(10),
                                  species=["H"] * 4 + ["He"] * 4,
                                  coords=[[0.0, 0.0, 0.0],
                                          [0.5, 0.5, 0.0],
                                          [0.5, 0.0, 0.5],
                                          [0.0, 0.5, 0.5],

                                          [0.0, 0.0, 0.5],
                                          [0.0, 0.5, 0.0],
                                          [0.5, 0.0, 0.0],
                                          [0.5, 0.5, 0.5],
                                        ])
    with pytest.raises(NotPrimitiveError):
        append_interstitial(cubic_supercell_info_wo_int, conventional_cell,
                            [[1/4, 1/4, 1/4]], ["test1"])




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from dataclasses import dataclass
from typing import Optional, Dict

from monty.json import MSONable
from pydefect.analyzer.defect_structure_info import SymmRelation, DefectType
from tabulate import tabulate
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class SingleCalcSummary(MSONable):
    charge: int
    atom_io: dict
    electronic_conv: Optional[bool] = None
    ionic_conv: Optional[bool] = None
    is_energy_strange: Optional[bool] = None
    same_config_from_init: Optional[bool] = None
    defect_type: Optional[str] = None
    symm_relation: Optional[str] = None
    donor_phs: Optional[bool] = None
    acceptor_phs: Optional[bool] = None
    unoccupied_deep_state: Optional[bool] = None
    occupied_deep_state: Optional[bool] = None
    same_structure: Optional[str] = None

    def same_atom_charge_io(self, other: "SingleCalcSummary"):
        return self.charge == other.charge and self.atom_io == other.atom_io

    @property
    def is_converged(self):
        return self.electronic_conv and self.ionic_conv

    @property
    def is_proper_result(self):
        return self.is_converged is True and self.is_energy_strange is False

    @property
    def is_unusual(self):
        return (self.symm_relation is SymmRelation.supergroup or
                self.same_config_from_init is False)

    @property
    def config_list(self):
        if self.is_proper_result:
            result = [".", ".", "."]
            if self.same_config_from_init in (False, None):
                result.extend([str(self.same_config_from_init),
                               str(self.defect_type),
                               str(self.symm_relation)])
            else:
                result.extend([".", ".", "."])
            return result
        elif self.is_converged:
            return [".", ".", self.is_energy_strange]
        else:
            return [self.electronic_conv, self.ionic_conv]


@dataclass
class CalcSummary(MSONable, ToJsonFileMixIn):
    single_summaries: Dict[str, SingleCalcSummary]

    def __str__(self):
        lines = [["name", "Ele. conv.", "Ionic conv.", "Is energy strange",
                  "Same config.", "Defect type", "Symm. Relation"]]
        for k, v in self.single_summaries.items():
            lines.append([k] + v.config_list)
        return tabulate(lines, stralign="center", tablefmt="pipe")



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from copy import copy

from pandas import DataFrame
from pandas._testing import assert_frame_equal
from pydefect.cli.vasp.make_local_extrema import extrema_coords, \
    find_inequivalent_coords, \
    make_local_extrema_from_volumetric_data
from pydefect.input_maker.local_extrema import CoordInfo
from pydefect.util.structure_tools import Coordination
from pymatgen.io.vasp import Chgcar, VolumetricData

import numpy as np


def test_make_local_extrema_from_volumetric_data(vasp_files, vol_params):
    aeccar0 = Chgcar.from_file(str(vasp_files / "NaMgF3_AECCAR0"))
    aeccar2 = Chgcar.from_file(str(vasp_files / "NaMgF3_AECCAR2"))
    aeccar = aeccar0 + aeccar2
    aeccar.write_file("CHGCAR")
    make_local_extrema_from_volumetric_data(aeccar, vol_params)


def test_find_inequivalent_coords(simple_cubic):

    df = DataFrame([[0.1, 0.0, 0.0, 1.0],
                    [0.0, 0.1, 0.0, 1.0],
                    [0.0, 0.0, 0.1, 1.0],
                    [0.9, 0.0, 0.0, 1.0],
                    [0.0, 0.9, 0.0, 1.0],
                    [0.0, 0.0, 0.9, 1.0]],
                   columns=["a", "b", "c", "value"])
    actual = find_inequivalent_coords(simple_cubic, df)
    expected = CoordInfo(
        site_symmetry="4mm",
        coordination=Coordination({"H": [0.1]}, 0.13, [0]),
        frac_coords=[(0.1, 0.0, 0.0), (0.0, 0.1, 0.0), (0.0, 0.0, 0.1),
                     (0.9, 0.0, 0.0), (0.0, 0.9, 0.0), (0.0, 0.0, 0.9)],
        quantities=[1.0]*6)
    assert actual[0] == expected


def test_extrema_coords(vasp_files, simple_cubic, vol_params):
    aeccar = VolumetricData(simple_cubic,
                            data={"total": np.array(
                                [[[0.0, 0.0], [-1.0, 0.0]],
                                 [[-1.0, 0.0], [-2.0, -1.0]]])})
    params = copy(vol_params)
    params.min_dist = 0.01
    params.radius = 0.51
    actual = extrema_coords(volumetric_data=aeccar, find_min=True, params=params)
    expected = DataFrame([[0.5, 0.5, 0.0, -2.0, -1.25]],
                         columns=["a", "b", "c", "value", "ave_value"])
    assert_frame_equal(actual, expected)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import pytest
from monty.serialization import loadfn

from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.unitcell import Unitcell
from pydefect.cli.vasp.make_efnv_correction import \
    make_efnv_correction
from pydefect.input_maker.defect_entry import DefectEntry


@pytest.mark.skip()
def test(vasp_files):
    d = vasp_files / "KInO2_Va_O_2"
    defect_entry: DefectEntry = loadfn(d / "Va_O1_2/defect_entry.json")
    calc_results: CalcResults = loadfn(d / "Va_O1_2/calc_results.json")
    perfect_calc_results: CalcResults = loadfn(d / "perfect_calc_results.json")
    unitcell: Unitcell = loadfn(d / "unitcell.json")

    make_efnv_correction(defect_entry.charge,
                         calc_results,
                         perfect_calc_results,
                         unitcell.dielectric_constant,
                         1)





# -*- coding: utf-8 -*-

import os
import sys
from math import sqrt
from pathlib import Path

import pytest
from pydefect.analyzer.defect_structure_comparator import SiteDiff
from pydefect.analyzer.defect_structure_info import Displacement, \
    DefectStructureInfo
from pydefect.corrections.efnv_correction import PotentialSite, \
    ExtendedFnvCorrection
from pydefect.input_maker.local_extrema import VolumetricDataAnalyzeParams
from pydefect.input_maker.supercell_info import Site, SupercellInfo, \
    Interstitial
from pymatgen.core import Lattice, IStructure, Structure

sys.path.append(os.path.join(os.path.dirname(__file__), 'helpers'))
# Need the following the share the fixture


@pytest.fixture(scope="session")
def vasp_files():
    return Path(__file__).parent / "cli" / "vasp" / "vasp_files"


@pytest.fixture(scope="session")
def simple_cubic():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def simple_cubic_2x1x1():
    lattice = Lattice.orthorhombic(2.0, 1.0, 1.0)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H", "H"], coords=coords)


@pytest.fixture(scope="session")
def simple_cubic_2x2x2():
    lattice = Lattice.cubic(2.0)
    coords = [[0.0, 0.0, 0.0],
              [0.5, 0.5, 0.0],
              [0.5, 0.0, 0.5],
              [0.0, 0.5, 0.5],
              [0.0, 0.0, 0.5],
              [0.0, 0.5, 0.0],
              [0.5, 0.0, 0.0],
              [0.5, 0.5, 0.5]]
    return IStructure(lattice=lattice, species=["H"] * 8, coords=coords)


@pytest.fixture(scope="session")
def monoclinic():
    lattice = Lattice.monoclinic(3, 4, 5, 100)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def complex_monoclinic():
    lattice = Lattice.monoclinic(3, 4, 5, 100)
    coords = [[0.0, 0.0, 0.0],
              [0.1, 0.0, 0.0],
              [0.9, 0.0, 0.0],
              [0.2, 0.0, 0.0],
              [0.8, 0.0, 0.0]]
    return IStructure(lattice=lattice,
                      species=["H", "He", "He", "He", "He"],
                      coords=coords)


@pytest.fixture(scope="session")
def rhombohedral():
    lattice = Lattice.rhombohedral(a=1, alpha=45)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def elongated_tetragonal():
    lattice = Lattice.tetragonal(a=1, c=3 * sqrt(2))
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def elongated_tetragonal():
    lattice = Lattice.tetragonal(a=1, c=3 * sqrt(2))
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def a_centered_orthorhombic():
    lattice = Lattice([[1,  0, 0],
                       [0,  2, 3],
                       [0, -2, 3]])
    coords = [[0.5, 0.8, 0.8],
              [0.0, 0.3, 0.0],
              [0.0, 0.0, 0.3]]

    return IStructure(lattice=lattice, species=["H"] * 3, coords=coords)


@pytest.fixture(scope="session")
def c_centered_monoclinic():
    lattice = Lattice.monoclinic(3, 4, 5, 100)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0]]
    return IStructure(lattice=lattice, species=["H", "H"], coords=coords)


@pytest.fixture(scope="session")
def fcc():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0],
              [0.5, 0.0, 0.5], [0.0, 0.5, 0.5]]
    return IStructure(lattice=lattice, species=["H"] * 4, coords=coords)


@pytest.fixture(scope="session")
def bcc():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]
    return IStructure(lattice=lattice, species=["H"] * 2, coords=coords)


@pytest.fixture(scope="session")
def tetra_close_to_cubic():
    lattice = Lattice.tetragonal(1.001 * 10 / sqrt(2), 10)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def ortho_conventional():
    lattice = Lattice.orthorhombic(5, 6, 7)
    coords = [
        [0.0, 0.0, 0.0],
        [0.5, 0.5, 0.0],
        [0.5, 0.0, 0.5],
        [0.0, 0.5, 0.5],

        [0.0, 0.0, 0.5],
        [0.0, 0.5, 0.0],
        [0.5, 0.0, 0.0],
        [0.5, 0.5, 0.5],
    ]
    return IStructure(lattice=lattice, species=["H"] * 4 + ["He"] * 4,
                      coords=coords)


@pytest.fixture(scope="session")
def cubic_supercell():
    lattice = Lattice.cubic(10)
    coords = [
       [0.00, 0.00, 0.00],
       [0.00, 0.00, 0.50],
       [0.00, 0.50, 0.00],
       [0.00, 0.50, 0.50],
       [0.50, 0.00, 0.00],
       [0.50, 0.00, 0.50],
       [0.50, 0.50, 0.00],
       [0.50, 0.50, 0.50],
       [0.00, 0.25, 0.25],
       [0.00, 0.25, 0.75],
       [0.00, 0.75, 0.25],
       [0.00, 0.75, 0.75],
       [0.50, 0.25, 0.25],
       [0.50, 0.25, 0.75],
       [0.50, 0.75, 0.25],
       [0.50, 0.75, 0.75],
       [0.25, 0.00, 0.25],
       [0.25, 0.00, 0.75],
       [0.25, 0.50, 0.25],
       [0.25, 0.50, 0.75],
       [0.75, 0.00, 0.25],
       [0.75, 0.00, 0.75],
       [0.75, 0.50, 0.25],
       [0.75, 0.50, 0.75],
       [0.25, 0.25, 0.00],
       [0.25, 0.25, 0.50],
       [0.25, 0.75, 0.00],
       [0.25, 0.75, 0.50],
       [0.75, 0.25, 0.00],
       [0.75, 0.25, 0.50],
       [0.75, 0.75, 0.00],
       [0.75, 0.75, 0.50],
       [0.25, 0.00, 0.00],
       [0.25, 0.00, 0.50],
       [0.25, 0.50, 0.00],
       [0.25, 0.50, 0.50],
       [0.75, 0.00, 0.00],
       [0.75, 0.00, 0.50],
       [0.75, 0.50, 0.00],
       [0.75, 0.50, 0.50],
       [0.25, 0.25, 0.25],
       [0.25, 0.25, 0.75],
       [0.25, 0.75, 0.25],
       [0.25, 0.75, 0.75],
       [0.75, 0.25, 0.25],
       [0.75, 0.25, 0.75],
       [0.75, 0.75, 0.25],
       [0.75, 0.75, 0.75],
       [0.00, 0.00, 0.25],
       [0.00, 0.00, 0.75],
       [0.00, 0.50, 0.25],
       [0.00, 0.50, 0.75],
       [0.50, 0.00, 0.25],
       [0.50, 0.00, 0.75],
       [0.50, 0.50, 0.25],
       [0.50, 0.50, 0.75],
       [0.00, 0.25, 0.00],
       [0.00, 0.25, 0.50],
       [0.00, 0.75, 0.00],
       [0.00, 0.75, 0.50],
       [0.50, 0.25, 0.00],
       [0.50, 0.25, 0.50],
       [0.50, 0.75, 0.00],
       [0.50, 0.75, 0.50],
    ]
    return Structure(lattice=lattice, species=["H"] * 32 + ["He"] * 32,
                     coords=coords)


@pytest.fixture(scope="session")
def cubic_supercell_w_vacancy(cubic_supercell):
    result = cubic_supercell.copy()
    result.remove_sites([0])
    return result


@pytest.fixture(scope="session")
def cubic_supercell_w_vacancy_w_perturb(cubic_supercell_w_vacancy):
    result = cubic_supercell_w_vacancy.copy()
    result.translate_sites([0], [0.0, 0.0, 0.1])
    return result


@pytest.fixture
def supercell_info(ortho_conventional):
    sites = {"H1": Site(element="H", wyckoff_letter="a", site_symmetry="mmm",
                        equivalent_atoms=[0, 1, 2, 3]),
             "He1": Site(element="He", wyckoff_letter="b", site_symmetry="mmm",
                         equivalent_atoms=[4, 5, 6, 7])}
    interstitial = Interstitial(frac_coords=[0.25]*3,
                                site_symmetry="yy", info="test")
    return SupercellInfo(ortho_conventional,
                         "Fmmm",
                         [[1, 0, 0], [0, 1, 0], [0, 0, 1]], sites,
                         [interstitial])


@pytest.fixture
def cubic_supercell_info_wo_int(cubic_supercell):
    sites = {"H1": Site(element="H", wyckoff_letter="a", site_symmetry="m-3m",
                        equivalent_atoms=[0, 1, 2, 3]),
             "He1": Site(element="He", wyckoff_letter="b", site_symmetry="m-3m",
                         equivalent_atoms=[4, 5, 6, 7])}
    unitcell = Structure(Lattice.rhombohedral(7.071068, 60),
                         species=["H", "He"],
                         coords=[[0.0]*3, [0.5]*3])
    return SupercellInfo(cubic_supercell,
                         "Fm-3m", [[-2, 2, 2], [2, -2, 2], [2, 2, -2]], sites,
                         unitcell_structure=unitcell)


@pytest.fixture
def efnv_correction():
    s1 = PotentialSite(specie="H", distance=1.999, potential=1.0,
                       pc_potential=None)
    s2 = PotentialSite(specie="He", distance=2.0001, potential=1.5,
                       pc_potential=0.2)
    s3 = PotentialSite(specie="He", distance=3.0, potential=2.0,
                       pc_potential=0.3)

    return ExtendedFnvCorrection(charge=10,
                                 point_charge_correction=1.0,
                                 defect_region_radius=2.0,
                                 sites=[s1, s2, s3],
                                 defect_coords=(0.0, 0.0, 0.0))


@pytest.fixture(scope="session")
def before_refine():
    return Structure.from_str(fmt="poscar", input_string="""Mg4 O3
1.00000000000000
5 0 0
0 5 0
0 0 5
Mg   O
4     3
Direct
0.0051896248240553  0.9835077947659414  0.9945137498637422
0.0047282952713914  0.4827940046010823  0.4942929782542009
0.5040349492352973  0.9821499237428384  0.4944941755970405
0.5058945352747628  0.4828206016032297  0.9940420309511140
0.5045613848356609  0.4811103128264023  0.4933877756337353
0.0013796816599694  0.9829379087234287  0.4953360299212051
0.0083465288988691  0.4848714537370853  0.9941122597789658""")


# refine with anchor_atom_index=1, anchor_atom_coords=np.array([0.0, 0.5, 0.5])
@pytest.fixture(scope="session")
def after_refine():
    return Structure.from_str(fmt="poscar", input_string="""Mg4 O3
1.00000000000000
5 0 0
0 5 0
0 0 5
Mg   O
4     3
Direct
0.0 0.0 0.0
0.0 0.5 0.5
0.5 0.0 0.5
0.5 0.5 0.0
0.5 0.5 0.5
0.0 0.0 0.5
0.0 0.5 0.0""")


@pytest.fixture(scope="session")
def vol_params():
    return VolumetricDataAnalyzeParams(None, None, 0.5, 0.5, 0.4)


@pytest.fixture
def structures():
    perfect = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Li", "U"],
        coords=[[0.25, 0.25, 0.25],
                [0.75, 0.75, 0.75],
                [0.5, 0.5, 0.5],
                [0, 0, 0]])

    initial = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Be", "U"],
        coords=[[0.25, 0.25, 0.25],
                [0.75, 0.75, 0.75],
                [0.5, 0.5, 0.5],
                [0.0, 0.0, 0.0]])

    final = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Be", "U"],
        coords=[[0.27, 0.25, 0.2501],
                [0.76, 0.75, 0.7501],
                [0.5, 0.5, 0.501],
                [0.0, 0.0, 0.0001]])
    return perfect, initial, final


@pytest.fixture
def displacements():
    return [None,
            Displacement(specie="He",
                         original_pos=(0.75, 0.75, 0.75),
                         final_pos=(0.76, 0.75, 0.75),
                         distance_from_defect=6.465437190360757,
                         disp_vector=(0.10000000000000009, 0.0, 0.0),
                         displace_distance=0.09999999999999964,
                         angle=124.9),
            Displacement(specie="Be",
                         original_pos=(0.5, 0.5, 0.5),
                         final_pos=(0.5, 0.5, 0.5009),
                         distance_from_defect=2.1355627039494767,
                         disp_vector=(0.0, 0.0, 0.009000000000000119),
                         displace_distance=0.009000000000000341,
                         angle=125.8),
            Displacement(specie="U",
                         original_pos=(0.0, 0.0, 0.0),
                         final_pos=(0.0, 0.0, 0.0),
                         distance_from_defect=6.525191802736529,
                         disp_vector=(0.0, 0.0, 0.0),
                         displace_distance=0.0,
                         angle=None)]


@pytest.fixture
def def_str_info(displacements):
    site_diff = SiteDiff(removed=[(0, 'H', (0.25, 0.25, 0.25))],
                         inserted=[(0, 'H', (0.27, 0.25, 0.25))],
                         removed_by_sub=[(2, 'Li', (0.5, 0.5, 0.5))],
                         inserted_by_sub=[(2, 'Be', (0.5, 0.5, 0.5009))])
    site_diff_from_init = SiteDiff(removed=[(0, 'H', (0.25, 0.25, 0.25))],
                                   inserted=[(0, 'H', (0.27, 0.25, 0.25))],
                                   removed_by_sub=[(2, 'Li', (0.5, 0.5, 0.5))],
                                   inserted_by_sub=[(2, 'Be', (0.5, 0.5, 0.5))])
    shifted_final = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Be", "U"],
        coords=[[0.27, 0.25, 0.25],
                [0.76, 0.75, 0.75],
                [0.5, 0.5, 0.5009],
                [0.0, 0.0, 0.0]])
    return DefectStructureInfo(shifted_final_structure=shifted_final,
                               initial_site_sym="3m",
                               final_site_sym="m",
                               site_diff=site_diff,
                               site_diff_from_initial=site_diff_from_init,
                               symprec=0.1,
                               dist_tol=0.2,
                               anchor_atom_idx=3,
                               neighbor_atom_indices=[0],
                               neighbor_cutoff_factor=1.2,
                               drift_vector=(0.0, 0.0, 0.0001),
                               drift_dist=0.001,
                               center=(0.38, 0.375, 0.375175),
                               displacements=displacements)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pydefect.analyzer.unitcell import Unitcell
from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties


def make_unitcell_from_vasp(vasprun_band: Vasprun,
                            outcar_band: Outcar,
                            outcar_dielectric_clamped: Outcar,
                            outcar_dielectric_ionic: Outcar,
                            system_name: str = None) -> Unitcell:
    name = (system_name or
            vasprun_band.final_structure.composition.reduced_formula)
    outcar_dielectric_clamped.read_lepsilon()
    outcar_dielectric_ionic.read_lepsilon_ionic()
    band_edge_properties = VaspBandEdgeProperties(vasprun_band, outcar_band)
    vbm, cbm = band_edge_properties.vbm_cbm

    return Unitcell(
        system=name,
        # vbm and cbm are <class 'numpy.float64'>.
        vbm=float(vbm), cbm=float(cbm),
        ele_dielectric_const=outcar_dielectric_clamped.dielectric_tensor,
        ion_dielectric_const=outcar_dielectric_ionic.dielectric_ionic_tensor)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest

from pydefect.input_maker.defect_set import DefectSet, screen_defect_set
from pydefect.input_maker.defect_set_maker import charge_set
from pydefect.input_maker.defect import SimpleDefect, Defect


simple_defects = {SimpleDefect(None, "O1", [1]), SimpleDefect("N", "O1", [0, 1])}


@pytest.fixture
def defect_set():
    return DefectSet(defects=simple_defects)


def test_defect_set(defect_set):
    assert SimpleDefect(None, "O1", [1]) in defect_set
    assert len(defect_set) == 2
    # test iterable
    for defect in defect_set:
        defect


def test_to_yaml(defect_set, tmpdir):
    tmpdir.chdir()
    defect_set.to_yaml("tmp.yaml")
    expected = """N_O1: [0, 1]
Va_O1: [1]
"""
    assert tmpdir.join("tmp.yaml").read() == expected


def test_from_file(tmpdir):
    tmpdir.chdir()
    tmpdir.join("tmp.yaml").write("""N_O1: [0, 1]
Va_O1: [1] 
""")
    defect_set = DefectSet.from_yaml("tmp.yaml")
    assert isinstance(defect_set, DefectSet)
    assert defect_set == simple_defects


def test_charge_set():
    assert charge_set(3) == [-1, 0, 1, 2, 3]
    assert charge_set(2) == [0, 1, 2]
    assert charge_set(-3) == [-3, -2, -1, 0, 1]
    assert charge_set(-2) == [-2, -1, 0]


def test_screen_defect_set(defect_set):
    actual = screen_defect_set(defect_set, ["N_O1_1"])
    expected = {SimpleDefect("N", "O1", [1])}
    assert actual == DefectSet(expected)

"""
TODO
- Dopant

DONE
- Set Defect names
- Write yaml files
"""




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np

from pydefect.analyzer.calc_results import CalcResults
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection, PotentialSite
from pydefect.corrections.ewald import Ewald
from pydefect.corrections.gkfo_correction import GkfoCorrection
from pydefect.defaults import defaults


def make_gkfo_correction(efnv_correction: ExtendedFnvCorrection,
                         additional_charge: int,
                         final_calc_results: CalcResults,
                         initial_calc_results: CalcResults,
                         diele_tensor: np.array,
                         ion_clamped_diele_tensor: np.array,
                         accuracy: float = defaults.ewald_accuracy,
                         unit_conversion: float = 180.95128169876497
                         ) -> GkfoCorrection:
    assert final_calc_results.structure == initial_calc_results.structure
    assert abs(additional_charge) == 1

    defect_coords = efnv_correction.defect_coords
    lattice = initial_calc_results.structure.lattice
    ewald_ele = Ewald(lattice.matrix, ion_clamped_diele_tensor,
                      accuracy=accuracy)
    defect_region_radius = efnv_correction.defect_region_radius

    pc_2nd_term = - ewald_ele.lattice_energy

    gkfo_sites = []
    for index, site in enumerate(initial_calc_results.structure):
        specie = site.specie
        dist, _ = lattice.get_distance_and_image(site.frac_coords, defect_coords)
        pot = (final_calc_results.potentials[index]
               - initial_calc_results.potentials[index])

        rel_coord = [x - y for x, y in zip(site.frac_coords, defect_coords)]
        if dist <= defect_region_radius:
            pc_potential = None
        else:
            pc_potential = (ewald_ele.atomic_site_potential(rel_coord)
                            * additional_charge * unit_conversion)

        gkfo_sites.append(PotentialSite(specie, dist, pot, pc_potential))

    return GkfoCorrection(
        init_efnv_correction=efnv_correction,
        additional_charge=additional_charge,
        pc_2nd_term=pc_2nd_term * unit_conversion,
        gkfo_sites=gkfo_sites,
        ave_dielectric_tensor=np.trace(diele_tensor) / 3,
        ave_electronic_dielectric_tensor=np.trace(ion_clamped_diele_tensor) / 3)






# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections import defaultdict
from itertools import zip_longest
from typing import List, Tuple

import numpy as np
from pydefect.analyzer.band_edge_states import BandEdgeOrbitalInfos, \
    PerfectBandEdgeState, BandEdgeStates, EdgeInfo, OrbitalInfo, \
    LocalizedOrbital, BandEdgeState
from pydefect.analyzer.defect_charge_info import DefectChargeInfo
from pydefect.defaults import defaults


def get_similar_orb_idx(orbs: List[OrbitalInfo],
                        edge_info: EdgeInfo,
                        localized_orbs: List[int] = None,
                        _reversed: bool = False) -> Tuple[int, float]:
    if _reversed:
        orbs = orbs[::-1]  # by band

    de = defaults.similar_energy_criterion

    if localized_orbs:
        if _reversed:
            localized_orbs = [len(orbs) - i for i in localized_orbs]

        def is_passed(idx):
            return idx <= max(localized_orbs) - 1
    else:
        def is_passed(idx):
            return False

    for i, orb in enumerate(orbs):
        if is_passed(i):
            continue

        if ((_reversed is False and orb.energy > edge_info.energy - de) or
                (_reversed and orb.energy - de < edge_info.energy)):

            orb_diff = orbital_diff(orb.orbitals, edge_info.orbitals)
            if orb_diff < defaults.similar_orb_criterion:
                orb_idx = len(orbs) - i - 1 if _reversed else i
                return orb_idx, orb_diff
    raise ValueError(f"Similar orbital to the edge are not found.\n"
                     f"Energy criterion: {defaults.similar_energy_criterion}\n"
                     f"Orbital criterion: {defaults.similar_orb_criterion}\n"
                     f"Try to lower the criterion.")


def get_localized_orbs(orb_info_by_spin: List[List[OrbitalInfo]],
                       loc_band_index_range: List[int],
                       lowest_band_idx: int,
                       weights: List[float]):
    start = loc_band_index_range[0] - lowest_band_idx
    end = loc_band_index_range[1] + 1 - lowest_band_idx
    orbs_by_band_by_kpt = np.array(orb_info_by_spin)[:, start: end].T

    result = []
    for i, orbs_by_kpt in enumerate(orbs_by_band_by_kpt, loc_band_index_range[0]):
        def get_ave(key):
            return sum([getattr(orb, key) * w for orb, w in zip(orbs_by_kpt, weights)])

        orbs = defaultdict(list)
        for orb, w in zip(orbs_by_kpt, weights):
            for elem, contrb in orb.orbitals.items():
                for y, c in enumerate(contrb):
                    try:
                        orbs[elem][y] += c
                    except IndexError:
                        orbs[elem].append(0.0)
                        orbs[elem][y] += c

        result.append(LocalizedOrbital(
            band_idx=i,
            ave_energy=get_ave("energy"),
            occupation=get_ave("occupation"),
            orbitals=dict(orbs),
            participation_ratio=get_ave("participation_ratio")))
    return result


def num_electron_in_cbm(orb_info_by_spin: List[List[OrbitalInfo]],
                        cbm_idx: int,
                        weights: List[float]):
    result = 0.0
    orbs_by_band_by_kpt = np.array(orb_info_by_spin)[:, cbm_idx:].T
    for orbs_by_kpt in orbs_by_band_by_kpt:
        result += sum([o.occupation * w for o, w in zip(orbs_by_kpt, weights)])
    return result


def num_hole_in_vbm(orb_info_by_spin: List[List[OrbitalInfo]],
                    vbm_idx: int,
                    weights: List[float]):
    result = 0.0
    orbs_by_band_by_kpt = np.array(orb_info_by_spin)[:, :vbm_idx + 1].T
    for orbs_by_kpt in orbs_by_band_by_kpt:
        result += sum([(1 - o.occupation) * w
                       for o, w in zip(orbs_by_kpt, weights)])
    return result


def make_band_edge_states(orbital_infos: BandEdgeOrbitalInfos,
                          p_edge_state: PerfectBandEdgeState,
                          defect_charge_info: DefectChargeInfo = None
                          ) -> BandEdgeStates:
    """Make BandEdgeStates by comparing orbital infos from perfect supercell.
    """
    vbm_k_idx = orbital_infos.kpt_idx(p_edge_state.vbm_info.kpt_coord)
    cbm_k_idx = orbital_infos.kpt_idx(p_edge_state.cbm_info.kpt_coord)
    p_vbm_info = p_edge_state.vbm_info
    p_cbm_info = p_edge_state.cbm_info

    states = []
    lowest_idx = orbital_infos.lowest_band_index
    for spin_idx, orb_info_by_spin in enumerate(orbital_infos.orbital_infos):
        if defect_charge_info:
            localized_orbs = defect_charge_info.localized_orbitals()[spin_idx]
            localized_orbs = [i - lowest_idx for i in localized_orbs]
        else:
            localized_orbs = None

        vbm_idx, vbm_diff = get_similar_orb_idx(orb_info_by_spin[vbm_k_idx],
                                                p_vbm_info, localized_orbs,
                                                _reversed=True)
        vbm_info = EdgeInfo(band_idx=vbm_idx + lowest_idx,
                            kpt_coord=p_vbm_info.kpt_coord,
                            orbital_info=orb_info_by_spin[vbm_k_idx][vbm_idx])

        cbm_idx, cbm_diff = get_similar_orb_idx(orb_info_by_spin[cbm_k_idx],
                                                p_cbm_info, localized_orbs)
        cbm_info = EdgeInfo(band_idx=cbm_idx + lowest_idx,
                            kpt_coord=p_cbm_info.kpt_coord,
                            orbital_info=orb_info_by_spin[cbm_k_idx][cbm_idx])

        loc_idx_range = [vbm_info.band_idx + 1, cbm_info.band_idx - 1]
        localized_orbs = get_localized_orbs(orb_info_by_spin,
                                            loc_idx_range,
                                            orbital_infos.lowest_band_index,
                                            orbital_infos.kpt_weights)

        vbm_hole = num_hole_in_vbm(orb_info_by_spin,
                                   vbm_idx=vbm_idx,
                                   weights=orbital_infos.kpt_weights)
        cbm_electron = num_electron_in_cbm(orb_info_by_spin,
                                           cbm_idx=cbm_idx,
                                           weights=orbital_infos.kpt_weights)

        states.append(BandEdgeState(vbm_info=vbm_info,
                                    cbm_info=cbm_info,
                                    vbm_orbital_diff=vbm_diff,
                                    cbm_orbital_diff=cbm_diff,
                                    localized_orbitals=localized_orbs,
                                    vbm_hole_occupation=vbm_hole,
                                    cbm_electron_occupation=cbm_electron))

    return BandEdgeStates(states=states)


def orbital_diff(orbital_1: dict, orbital_2: dict) -> float:
    element_set = set(list(orbital_1.keys()) + list(orbital_2.keys()))
    orb_1, orb_2 = defaultdict(list, orbital_1), defaultdict(list, orbital_2)
    result = 0
    for e in element_set:
        result += sum([abs(i - j) for i, j
                       in zip_longest(orb_1[e], orb_2[e], fillvalue=0)])

    return result





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import numpy as np
import pytest
from pydefect.analyzer.defect_charge_info import DefectChargeInfo, AveChargeDensityDist
from pymatgen.electronic_structure.core import Spin
from pymatgen.core import Structure, Lattice
from pymatgen.io.vasp import Chgcar
from vise.tests.helpers.assertion import assert_json_roundtrip


@pytest.fixture
def ave_charge_density_dist():
    return AveChargeDensityDist(charge_center=(0.1, 0.1, 0.1),
                                radial_dist=[0.4, 0.4, 0.2])


@pytest.fixture
def defect_charge_info(ave_charge_density_dist):
    ave_charge_density_dist_down = AveChargeDensityDist(
        charge_center=(0.1, 0.1, 0.1), radial_dist=[0.3, 0.5, 0.2])

    return DefectChargeInfo(distance_bins=[0.0, 1.0, 2.0, 2.5],
                            band_idxs=[10],
                            charge_dists=[[ave_charge_density_dist,
                                           ave_charge_density_dist_down]],
                            ave_charge_density=0.5)


def test_charge_dist_json_roundtrip(ave_charge_density_dist, tmpdir):
    assert_json_roundtrip(ave_charge_density_dist, tmpdir)


def test_defect_charge_info_json_roundtrip(defect_charge_info, tmpdir):
    assert_json_roundtrip(defect_charge_info, tmpdir)


def test_defect_charge_info_sum_chg_dens_distribution(defect_charge_info):
    actual = defect_charge_info.sum_chg_dens_distribution(10, Spin.up)
    expected = np.array([0.4 * 1.0 ** 3,
                         0.4 * (2.0 ** 3 - 1.0 ** 3),
                         0.2 * (2.5 ** 3 - 2.0 ** 3)]) * 4 / 3 * np.pi
    np.testing.assert_array_almost_equal(actual, expected)


def test_defect_charge_info_half_charge_radius(defect_charge_info):
    actual = defect_charge_info.half_charge_radius(10, Spin.up)
    assert actual == 0.2984155182973037


def test_defect_charge_info_uniform_half_charge_radius(defect_charge_info):
    assert defect_charge_info.uniform_half_charge_radius == 0.6203504908994001


def test_defect_charge_info_repr(defect_charge_info):
    print(defect_charge_info)
    expected = """ -- defect charge info
Uniform charge radius is  0.620
Band index  Spin  Radius  Center
11          up    0.298   ( 0.100,  0.100,  0.100)
11          down  0.398   ( 0.100,  0.100,  0.100)"""
    assert defect_charge_info.__str__() == expected


def test_defect_charge_info_show_dist(defect_charge_info):
    plt = defect_charge_info.show_dist()
    plt.show()


def test_defect_charge_info_are_localized(defect_charge_info):
    actual = defect_charge_info.localized_orbitals(0.3, 1.0)
    expected = [[10], []]
    assert actual == expected

    actual = defect_charge_info.localized_orbitals(0.4, 1.0)
    expected = [[10], [10]]
    assert actual == expected

    actual = defect_charge_info.localized_orbitals(0.4, 0.398 / 0.62 - 0.001)
    expected = [[10], []]
    assert actual == expected





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from itertools import groupby
from typing import List, Dict

from pydefect.analyzer.band_edge_states import PerfectBandEdgeState
from pydefect.analyzer.defect_energy import DefectEnergyInfo, \
    DefectEnergySummary, DefectEnergies
from pydefect.analyzer.unitcell import Unitcell
from pydefect.chem_pot_diag.chem_pot_diag import TargetVertices


def make_defect_energy_summary(
        energy_infos: List[DefectEnergyInfo],
        target_vertices: TargetVertices,
        unitcell: Unitcell,
        perfect_band_edge: PerfectBandEdgeState) -> DefectEnergySummary:

    defect_energies = {}
    key = lambda x: x.name
    # MUST NEED SORTED.
    for _, grouped_es in groupby(sorted(energy_infos, key=key), key=key):
        grouped_es = list(grouped_es)
        name, atom_io = grouped_es[0].name, grouped_es[0].atom_io
        charges, des = [], []
        for es in grouped_es:
            assert atom_io == es.atom_io
            charges.append(es.charge)
            des.append(es.defect_energy)
        defect_energies[name] = DefectEnergies(atom_io, charges, des)

    return DefectEnergySummary(title=unitcell.system,
                               defect_energies=defect_energies,
                               rel_chem_pots=target_vertices.chem_pots,
                               cbm=unitcell.cbm - unitcell.vbm,
                               supercell_vbm=perfect_band_edge.vbm_info.energy - unitcell.vbm,
                               supercell_cbm=perfect_band_edge.cbm_info.energy - unitcell.vbm)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

from monty.serialization import loadfn
from pydefect.analyzer.eigenvalue_plotter import EigenvalueMplPlotter
from pydefect.chem_pot_diag.chem_pot_diag import CompositionEnergy, \
    CompositionEnergies
from pydefect.cli.main_tools import parse_dirs
from pydefect.cli.vasp.make_local_extrema import \
    make_local_extrema_from_volumetric_data
from pydefect.cli.vasp.make_band_edge_orbital_infos import \
    make_band_edge_orbital_infos
from pydefect.cli.vasp.make_calc_results import make_calc_results_from_vasp
from pydefect.cli.vasp.make_perfect_band_edge_state import \
    make_perfect_band_edge_state_from_vasp
from pydefect.cli.vasp.make_poscars_from_query import make_poscars_from_query
from pydefect.cli.vasp.make_unitcell import make_unitcell_from_vasp
from pydefect.input_maker.defect_entries_maker import DefectEntriesMaker
from pydefect.input_maker.defect_set import DefectSet
from pydefect.input_maker.local_extrema import VolumetricDataAnalyzeParams
from pydefect.input_maker.supercell_info import SupercellInfo
from pydefect.util.mp_tools import MpQuery
from pymatgen.core import Structure
from pymatgen.io.vasp import Vasprun, Outcar, Procar
from vise.defaults import defaults
from vise.util.logger import get_logger

logger = get_logger(__name__)


def make_unitcell(args):
    unitcell = make_unitcell_from_vasp(
        vasprun_band=args.vasprun_band,
        outcar_band=args.outcar_band,
        outcar_dielectric_clamped=args.outcar_dielectric_clamped,
        outcar_dielectric_ionic=args.outcar_dielectric_ionic,
        system_name=args.name)
    unitcell.to_yaml_file()


def make_competing_phase_dirs(args):
    query = MpQuery(element_list=args.elements, e_above_hull=args.e_above_hull)
    make_poscars_from_query(materials_query=query.materials, path=Path.cwd())


def make_composition_energies(args):
    if args.yaml_file:
        composition_energies = CompositionEnergies.from_yaml(args.yaml_file)
    else:
        composition_energies = CompositionEnergies()

    def _inner(_dir: Path):
        outcar = Outcar(_dir / defaults.outcar)
        composition = Structure.from_file(_dir / defaults.contcar).composition
        energy = float(outcar.final_energy)  # original type is FloatWithUnit
        return composition, CompositionEnergy(energy, "local")

    for c, ce in parse_dirs(args.dirs, _inner, args.verbose):
        composition_energies[c] = ce

    composition_energies.to_yaml_file()


def make_local_extrema(args):
    volumetric_data = args.volumetric_data[0]
    if len(args.volumetric_data) > 1:
        for i in args.volumetric_data[1:]:
            volumetric_data += i

    params = VolumetricDataAnalyzeParams(args.threshold_frac,
                                         args.threshold_abs,
                                         args.min_dist,
                                         args.tol,
                                         args.radius)
    local_extrema = make_local_extrema_from_volumetric_data(
        volumetric_data=volumetric_data,
        params=params,
        info=args.info,
        find_min=not args.find_max)
    local_extrema.to_json_file()


def make_defect_entries(args):
    supercell_info: SupercellInfo = loadfn("supercell_info.json")
    perfect = Path("perfect")

    try:
        perfect.mkdir()
        logger.info("Making perfect dir...")
        supercell_info.structure.to(filename=perfect / "POSCAR")
    except FileExistsError:
        logger.info(f"perfect dir exists, so skipped...")

    defect_set = DefectSet.from_yaml()
    maker = DefectEntriesMaker(supercell_info, defect_set)

    for defect_entry in maker.defect_entries:
        dir_path = Path(defect_entry.full_name)
        try:
            dir_path.mkdir()
            logger.info(f"Making {dir_path} dir...")
            if defect_entry.perturbed_structure:
                defect_entry.perturbed_structure.to(filename=dir_path / "POSCAR")
            else:
                defect_entry.structure.to(filename=dir_path / "POSCAR")

            defect_entry.to_json_file(filename=dir_path / "defect_entry.json")
            defect_entry.to_prior_info(filename=dir_path / "prior_info.yaml")
        except FileExistsError:
            logger.info(f"{dir_path} dir exists, so skipped...")


def make_calc_results(args):
    file_name = "calc_results.json"

    def _inner(_dir: Path):
        calc_results = make_calc_results_from_vasp(
            vasprun=Vasprun(_dir / defaults.vasprun),
            outcar=Outcar(_dir / defaults.outcar))
        calc_results.to_json_file(str(_dir / file_name))

    parse_dirs(args.dirs, _inner, args.verbose, file_name)


def make_perfect_band_edge_state(args):
    procar = Procar(args.dir / defaults.procar)
    vasprun = Vasprun(args.dir / defaults.vasprun)
    outcar = Outcar(args.dir / defaults.outcar)
    perfect_band_edge_state = \
        make_perfect_band_edge_state_from_vasp(procar, vasprun, outcar)
    perfect_band_edge_state.to_json_file(
        args.dir / "perfect_band_edge_state.json")


def make_band_edge_orb_infos_and_eigval_plot(args):
    supercell_vbm = args.p_state.vbm_info.energy
    supercell_cbm = args.p_state.cbm_info.energy

    file_name = "band_edge_orbital_infos.json"
    def _inner(_dir: Path):
        try:
            defect_entry = loadfn(_dir / "defect_entry.json")
            title = defect_entry.name
        except FileNotFoundError:
            title = "No name"
        procar = Procar(_dir / defaults.procar)
        vasprun = Vasprun(_dir / defaults.vasprun)
        str_info = loadfn(_dir / "defect_structure_info.json")
        band_edge_orb_chars = make_band_edge_orbital_infos(
            procar, vasprun, supercell_vbm, supercell_cbm, str_info)
        band_edge_orb_chars.to_json_file(_dir / file_name)

        plotter = EigenvalueMplPlotter(
            title=title, band_edge_orb_infos=band_edge_orb_chars,
            supercell_vbm=supercell_vbm, supercell_cbm=supercell_cbm,
            y_range=args.y_range)
        plotter.construct_plot()
        plotter.plt.savefig(fname=_dir / "eigenvalues.pdf")
        plotter.plt.clf()

    parse_dirs(args.dirs, _inner, args.verbose, file_name)